import os

import pytest
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS as AstropyWCS

from .helpers import generate_file, generate_hdu, generate_wcs
from .. import FITSFigure

# The tests in this file check that the initialization and basic plotting do
# not crash for FITS files with 2 dimensions. No reference images are
# required here.

header_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data/2d_fits')

HEADERS = [os.path.join(header_dir, '1904-66_AIR.hdr'),
           os.path.join(header_dir, '1904-66_AIT.hdr'),
           os.path.join(header_dir, '1904-66_ARC.hdr'),
           os.path.join(header_dir, '1904-66_AZP.hdr'),
           os.path.join(header_dir, '1904-66_BON.hdr'),
           os.path.join(header_dir, '1904-66_CAR.hdr'),
           os.path.join(header_dir, '1904-66_CEA.hdr'),
           os.path.join(header_dir, '1904-66_COD.hdr'),
           os.path.join(header_dir, '1904-66_COE.hdr'),
           os.path.join(header_dir, '1904-66_COO.hdr'),
           os.path.join(header_dir, '1904-66_COP.hdr'),
           os.path.join(header_dir, '1904-66_CSC.hdr'),
           os.path.join(header_dir, '1904-66_CYP.hdr'),
           os.path.join(header_dir, '1904-66_HPX.hdr'),
           os.path.join(header_dir, '1904-66_MER.hdr'),
           os.path.join(header_dir, '1904-66_MOL.hdr'),
           os.path.join(header_dir, '1904-66_NCP.hdr'),
           os.path.join(header_dir, '1904-66_PAR.hdr'),
           os.path.join(header_dir, '1904-66_PCO.hdr'),
           os.path.join(header_dir, '1904-66_QSC.hdr'),
           os.path.join(header_dir, '1904-66_SFL.hdr'),
           os.path.join(header_dir, '1904-66_SIN.hdr'),
           os.path.join(header_dir, '1904-66_STG.hdr'),
           os.path.join(header_dir, '1904-66_SZP.hdr'),
           os.path.join(header_dir, '1904-66_TAN.hdr'),
           os.path.join(header_dir, '1904-66_TSC.hdr'),
           os.path.join(header_dir, '1904-66_ZEA.hdr'),
           os.path.join(header_dir, '1904-66_ZPN.hdr')]

REFERENCE = os.path.join(header_dir, '1904-66_TAN.hdr')

CAR_REFERENCE = os.path.join(header_dir, '1904-66_CAR.hdr')

VALID_DIMENSIONS = [(0, 1), (1, 0)]
INVALID_DIMENSIONS = [None, (1,), (0, 2), (-4, 2), (1, 1), (2, 2), (1, 2, 3)]


# Test initialization through a filename
def test_file_init(tmpdir):
    filename = generate_file(REFERENCE, str(tmpdir))
    f = FITSFigure(filename)
    f.show_grayscale()
    f.close()


# Test initialization through an HDU object
def test_hdu_init():
    hdu = generate_hdu(REFERENCE)
    f = FITSFigure(hdu)
    f.show_grayscale()
    f.close()


# Test initialization through a WCS object
def test_wcs_init():
    wcs = generate_wcs(REFERENCE)
    f = FITSFigure(wcs)
    f.show_grayscale()
    f.close()


# Test initialization through a WCS object with wcs.to_header() as a go-between
# specifically for testing the cd -> pc -> cd hack, and has particular importance
# for AVM-generated headers
def test_wcs_toheader_init():
    wcs = generate_wcs(REFERENCE)
    header_ = fits.Header.fromtextfile(REFERENCE)
    header = wcs.to_header()
    wcs2 = AstropyWCS(header)
    wcs2.pixel_shape = wcs.pixel_shape = (header_['NAXIS1'], header_['NAXIS2'])
    f = FITSFigure(wcs2)
    f.show_grayscale()
    f.add_grid()
    f.close()


# Test initialization through an HDU object (no WCS)
def test_hdu_nowcs_init():
    data = np.zeros((16, 16))
    hdu = fits.PrimaryHDU(data)
    f = FITSFigure(hdu)
    f.show_grayscale()
    f.close()


# Test initalization through a Numpy array (no WCS)
def test_numpy_nowcs_init():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.show_grayscale()
    f.close()


# Now check initialization with valid and invalid dimensions. We just need to
# tes with HDU objects since we already tested that reading from files is ok.

# Test initialization with valid dimensions
@pytest.mark.parametrize(('dimensions'), VALID_DIMENSIONS)
def test_init_dimensions_valid(dimensions):
    hdu = generate_hdu(REFERENCE)
    f = FITSFigure(hdu, dimensions=dimensions)
    f.show_grayscale()
    f.close()

# Test initialization with invalid dimensions


@pytest.mark.parametrize(('dimensions'), INVALID_DIMENSIONS)
def test_init_dimensions_invalid(dimensions):
    hdu = generate_hdu(REFERENCE)
    with pytest.raises(ValueError):
        FITSFigure(hdu, dimensions=dimensions)

# Now check initialization of different WCS projections, and we check only
# valid dimensions


valid_parameters = []
for h in HEADERS:
    for d in VALID_DIMENSIONS:
        valid_parameters.append((h, d))


@pytest.mark.parametrize(('header', 'dimensions'), valid_parameters)
def test_init_extensive_wcs(header, dimensions):
    hdu = generate_hdu(header)
    if 'CAR' in header:
        f = FITSFigure(hdu, dimensions=dimensions, convention='calabretta')
    else:
        f = FITSFigure(hdu, dimensions=dimensions)
    f.show_grayscale()
    f.add_grid()
    f.close()

# Check that for CAR projections, an exception is raised if no convention is specified


@pytest.mark.parametrize(('dimensions'), VALID_DIMENSIONS)
def test_init_car_invalid(dimensions):
    hdu = generate_hdu(CAR_REFERENCE)
    with pytest.raises(Exception):
        FITSFigure(hdu, dimensions=dimensions)

# Check that images containing only NaN or Inf values don't crash FITSFigure


def test_init_only_naninf():
    data = np.ones((10, 10)) * np.nan
    data[::2, ::2] = np.inf
    f = FITSFigure(data)
    f.show_grayscale()
    f.show_colorscale()


def test_init_single_pixel():
    data = np.zeros((4, 4))
    data[...] = np.nan
    data[2, 2] = 1
    f = FITSFigure(data)
    f.show_grayscale()


def test_not_first_hdu(tmpdir):

    # Test that data is fetched from the first compatible HDU with data

    filename = tmpdir.join('test.fits').strpath

    hdu0 = fits.PrimaryHDU()
    hdu1 = fits.BinTableHDU()
    hdu2 = fits.ImageHDU(np.zeros((12, 12)))

    hdulist = fits.HDUList([hdu0, hdu1, hdu2])

    hdulist.writeto(filename)

    f = FITSFigure(filename)
    assert f._data.shape == (12, 12)



import os
import tempfile

import numpy as np

from .. import FITSFigure
from .helpers import generate_file
from .figures import figure_test

MODULEDIR = os.path.dirname(__file__)
DATADIR = os.path.abspath(os.path.join(MODULEDIR, 'data'))


class BaseImageTests(object):

    @classmethod
    def setup_class(cls):

        cls._baseline_images_dir = os.path.abspath(os.path.join(MODULEDIR, 'baseline_images'))

        header_1 = os.path.join(DATADIR, '2d_fits/1904-66_AIR.hdr')
        cls.filename_1 = generate_file(header_1, str(tempfile.mkdtemp()))

        header_2 = os.path.join(DATADIR, '2d_fits/2MASS_k.hdr')
        cls.filename_2 = generate_file(header_2, str(tempfile.mkdtemp()))

        header_3 = os.path.join(DATADIR, '3d_fits/cube.hdr')
        cls.filename_3 = generate_file(header_3, str(tempfile.mkdtemp()))

        header_4 = os.path.join(DATADIR, '2d_fits/2MASS_k_rot.hdr')
        cls.filename_4 = generate_file(header_4, str(tempfile.mkdtemp()))


class TestBasic(BaseImageTests):

    # Test for showing grayscale
    @figure_test
    def test_basic_image(self):
        f = FITSFigure(self.filename_2, figsize=(7, 5))
        f.show_grayscale(vmin=0, vmax=1)
        return f

    @figure_test
    def test_ticks_labels_options(self):
        f = FITSFigure(self.filename_2, figsize=(7, 5))

        # Force aspect ratio
        f.show_grayscale()
        f.hide_grayscale()

        f.ticks.set_color('black')
        f.axis_labels.set_xposition('top')
        f.axis_labels.set_yposition('right')
        f.axis_labels.set_font(size='medium', weight='medium',
                               stretch='normal', style='normal')
        f.tick_labels.set_xformat('dd:mm:ss.ss')
        f.tick_labels.set_yformat('hh:mm:ss.ss')
        f.tick_labels.set_style('colons')
        f.ticks.set_xspacing(0.2)
        f.ticks.set_yspacing(0.2)
        f.ticks.set_minor_frequency(10)
        return f

    # Test for showing colorscale
    @figure_test
    def test_show_colorbar_scalebar_beam(self):
        f = FITSFigure(self.filename_1, figsize=(7, 5))
        f.ticks.set_color('black')
        f.show_colorscale(vmin=-0.1, vmax=0.1)
        f.add_colorbar()
        f.add_scalebar(7.5)
        f.add_beam(major=0.5, minor=0.2, angle=10.)
        f.tick_labels.hide()
        return f

    # Test for overlaying shapes
    @figure_test
    def test_overlay_shapes(self):
        f = FITSFigure(self.filename_1, figsize=(7, 5))

        # Force aspect ratio
        f.show_grayscale()
        f.hide_grayscale()

        f.ticks.set_color('black')

        # Markers
        f.show_markers([360., 350., 340.], [-61., -62., -63], color='cyan')
        f.show_markers([30, 40], [50, 70], coords_frame='pixel', edgecolor='blue')

        # Circles
        f.show_circles([360., 350., 340.], [-61., -62., -63], [0.5, 0.4, 0.3], edgecolor='purple')
        f.show_circles([30, 40], [50, 70], [10, 20], coords_frame='pixel', edgecolor='orange')

        # Ellipses
        f.show_ellipses(340., -66., 1.5, 2., 10., edgecolor='red')
        f.show_ellipses(120, 60, 20, 40, 20., coords_frame='pixel', edgecolor='green')

        # Rectangles
        f.show_rectangles([355., 350.], [-71, -72], [0.5, 1], [2, 1], angle=[20, 30], edgecolor='magenta')
        f.show_rectangles([66, 80], [20, 30], [10, 14], [20, 22], angle=[20, 30], coords_frame='pixel', edgecolor='yellow')

        # Arrows
        f.show_arrows([340., 360], [-72, -68], [2, -2], [2, 2])

        # Polygons
        poly = np.array([[330, 340, 360], [-65, -61, -63]])
        f.show_polygons([poly], edgecolor='0.3', zorder=10)

        # Lines
        f.show_lines([poly], zorder=9, lw=5, color='red', alpha=0.5)

        # Labels
        f.add_label(350., -66., 'text')
        f.add_label(0.4, 0.25, 'text', relative=True)

        f.frame.set_linewidth(1)  # points
        f.frame.set_color('black')
        f.axis_labels.hide()

        return f

    # Test for grid
    @figure_test
    def test_grid(self):
        f = FITSFigure(self.filename_1, figsize=(7, 5))

        # Force aspect ratio
        f.show_grayscale()
        f.hide_grayscale()

        f.ticks.set_color('black')
        f.add_grid()
        f.grid.set_color('red')
        f.grid.set_alpha(0.8)
        f.grid.set_linestyle('solid')
        f.grid.set_xspacing('tick')
        f.grid.set_yspacing(3)
        return f

    # Test recenter
    @figure_test
    def test_recenter(self):
        f = FITSFigure(self.filename_2, figsize=(7, 5))

        # Force aspect ratio
        f.show_grayscale()
        f.hide_grayscale()

        f.ticks.set_color('black')
        f.recenter(266.5, -29.0, width=0.1, height=0.1)
        f.axis_labels.set_xpad(20)
        f.axis_labels.set_ypad(20)
        return f

    # Test overlaying contours
    @figure_test
    def test_contours(self):
        data = np.arange(256).reshape((16, 16))
        f = FITSFigure(data, figsize=(7, 5))

        # Force aspect ratio
        f.show_grayscale()
        f.hide_grayscale()

        f.ticks.set_color('black')
        f.show_contour(data, levels=np.linspace(1., 254., 10), filled=False)
        return f

    # Test cube slice
    @figure_test
    def test_cube_slice(self):
        f = FITSFigure(self.filename_3, dimensions=[2, 0], slices=[10], figsize=(7, 5), subplot=[0.25, 0.1, 0.7, 0.8])
        f.ticks.set_color('black')
        f.add_grid()
        f.grid.set_color('black')
        f.grid.set_linestyle('solid')
        f.grid.set_xspacing(250)
        f.grid.set_yspacing(0.01)
        f.tick_labels.set_xformat('%g')
        f.tick_labels.set_yformat('dd:mm:ss.ss')
        return f

    # Test for ds9 regions
    @figure_test
    def test_regions(self):
        f = FITSFigure(self.filename_2, figsize=(7, 5))

        # Force aspect ratio
        f.show_grayscale()
        f.hide_grayscale()

        f.show_regions(os.path.join(DATADIR, 'shapes.reg'))
        f.axis_labels.hide()
        f.tick_labels.hide()
        f.ticks.hide()
        return f

    @figure_test
    def test_north(self):
        f = FITSFigure(self.filename_4, figsize=(3, 3), north=True)
        f.show_grayscale(vmin=-1, vmax=1)
        f.axis_labels.hide()
        f.tick_labels.hide()
        f.ticks.hide()
        return f

    @figure_test
    def test_downsample(self):
        data = np.arange(256).reshape((16, 16))
        f = FITSFigure(data, downsample=2)
        f.show_grayscale()
        return f

    @figure_test
    def test_set_nan_color(self):
        data = np.arange(56, dtype=float).reshape((8, 7))
        data[3, :] = np.nan
        f = FITSFigure(data, figsize=(3, 3))
        f.show_colorscale()
        f.axis_labels.hide()
        f.tick_labels.hide()
        f.ticks.hide()
        f.set_nan_color('black')
        return f



import os
import operator
from functools import reduce

import matplotlib

import matplotlib.pyplot as plt
from matplotlib.patches import Circle, Rectangle, Ellipse, Polygon, FancyArrow
from matplotlib.collections import PatchCollection, LineCollection

import numpy as np

from astropy import log
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from astropy.io import fits

try:
    from astropy.nddata.blocks import block_reduce
except ImportError:  # astropy<5
    from astropy.nddata.utils import block_reduce

from astropy.visualization import AsymmetricPercentileInterval
from astropy.visualization.wcsaxes import WCSAxes, WCSAxesSubplot
from astropy.coordinates import ICRS

from . import convolve_util
from . import header as header_util
from . import slicer

from .compat import simple_norm
from .layers import Layers
from .grid import Grid
from .ticks import Ticks
from .tick_labels import TickLabels
from .axis_labels import AxisLabels
from .overlays import Beam, Scalebar
from .regions import Regions
from .colorbar import Colorbar
from .frame import Frame

from .decorators import auto_refresh, fixdocstring

HDU_TYPES = tuple([fits.PrimaryHDU, fits.ImageHDU, fits.CompImageHDU])


__doctest_skip__ = ['FITSFigure.add_beam', 'FITSFigure.add_colorbar',
                    'FITSFigure.add_grid', 'FITSFigure.add_scalebar']


def uniformize_1d(*args):
    if len(args) > 1:
        return np.broadcast_arrays(np.atleast_1d(args[0]), *args[1:])
    elif len(args) == 1:
        return np.atleast_1d(args[0])
    else:
        raise ValueError("No arguments passed to uniformize_1d")


class FITSFigure(Layers, Regions):
    """
    Create a FITSFigure instance.

    This class is a wrapper around the Astropy WCSAxes class and provides
    the same API as historical versions of APLpy.

    Parameters
    ----------

    data : see below

        The FITS file to open. The following data types can be passed:

             string
             astropy.io.fits.PrimaryHDU
             astropy.io.fits.ImageHDU
             astropy.wcs.WCS
             np.ndarray
             RGB image with AVM meta-data

    hdu : int, optional
        By default, the image in the primary HDU is read in. If a
        different HDU is required, use this argument.

    figure : ~matplotlib.figure.Figure, optional
        If specified, a subplot will be added to this existing
        matplotlib figure() instance, rather than a new figure
        being created from scratch.

    subplot : tuple or list, optional
        If specified, a subplot will be added at this position. If a tuple
        of three values, the tuple should contain the standard matplotlib
        subplot parameters, i.e. (ny, nx, subplot). If a list of four
        values, the list should contain [xmin, ymin, dx, dy] where xmin
        and ymin are the position of the bottom left corner of the
        subplot, and dx and dy are the width and height of the subplot
        respectively. These should all be given in units of the figure
        width and height. For example, [0.1, 0.1, 0.8, 0.8] will almost
        fill the entire figure, leaving a 10 percent margin on all sides.

    downsample : int, optional
        If this option is specified, the image will be downsampled
        by a factor *downsample* when reading in the data.

    north : bool, optional
        Whether to rotate the image so that north is up. By default, this is
        assumed to be 'north' in the ICRS frame, but you can also pass any
        astropy :class:`~astropy.coordinates.BaseCoordinateFrame` to indicate
        to use the north of that frame.

    convention : str, optional
        This is used in cases where a FITS header can be interpreted
        in multiple ways. For example, for files with a -CAR
        projection and CRVAL2=0, this can be set to 'wells' or
        'calabretta' to choose the appropriate convention.

    dimensions : tuple or list, optional
        The index of the axes to use if the data has more than three
        dimensions.

    slices : tuple or list, optional
        If a FITS file with more than two dimensions is specified,
        then these are the slices to extract. If all extra dimensions
        only have size 1, then this is not required.

    auto_refresh : bool, optional
        Whether to refresh the figure automatically every time a
        plotting method is called. This can also be set using the
        set_auto_refresh method. This defaults to `True` if and only if
        APLpy is being used from IPython and the Matplotlib backend is
        interactive.

    kwargs
        Any additional arguments are passed on to matplotlib's Figure()
        class. For example, to set the figure size, use the
        figsize=(xsize, ysize) argument (where xsize and ysize are in
        inches). For more information on these additional arguments, see
        the *Optional keyword arguments* section in the documentation for
        :class:`~matplotlib.figure.Figure`.
    """

    @auto_refresh
    def __init__(self, data, hdu=0, figure=None, subplot=(1, 1, 1),
                 downsample=False, north=False, convention=None,
                 dimensions=[0, 1], slices=[], auto_refresh=None,
                 **kwargs):

        self._wcsaxes_slices = ('x', 'y')

        if 'figsize' not in kwargs:
            kwargs['figsize'] = (10, 9)

        # Set the grid type
        if len(slices) > 0:
            self.grid_type = 'contours'
        else:
            self.grid_type = 'lines'

        if (isinstance(data, str) and
                data.split('.')[-1].lower() in ['png', 'jpg', 'tif']):

            from PIL import Image
            from pyavm import AVM

            # Remember image filename
            self._rgb_image = data

            # Find image size
            nx, ny = Image.open(data).size

            # Now convert AVM information to WCS
            data = AVM.from_image(data).to_wcs()

            # Need to scale CDELT values sometimes the AVM meta-data is only
            # really valid for the full-resolution image
            data.wcs.cdelt = [data.wcs.cdelt[0] * nx / float(nx), data.wcs.cdelt[1] * ny / float(ny)]
            data.wcs.crpix = [data.wcs.crpix[0] / nx * float(nx), data.wcs.crpix[1] / ny * float(ny)]

            # Update the NAXIS values with the true dimensions of the RGB image
            data.nx = nx
            data.ny = ny
            data.pixel_shape = (nx, ny)

        if isinstance(data, WCS):

            wcs = data

            if wcs.naxis != 2:
                raise ValueError("FITSFigure initialization via WCS objects "
                                 "can only be done with 2-dimensional WCS "
                                 "objects")

            if wcs.pixel_shape is None:
                raise ValueError("The WCS object does not contain any size "
                                 "information")

            header = wcs.to_header()
            header['NAXIS1'], header['NAXIS2'] = wcs.pixel_shape
            nx = header['NAXIS%i' % (dimensions[0] + 1)]
            ny = header['NAXIS%i' % (dimensions[1] + 1)]
            self._data = np.zeros((ny, nx), dtype=float)
            self._header = header
            self._wcs = WCS(header, relax=True)
            self._wcs.nx = nx
            self._wcs.ny = ny

            if downsample:
                log.warning("downsample argument is ignored if data "
                            "passed is a WCS object")
                downsample = False

            if north:
                log.warning("north argument is ignored if data "
                            "passed is a WCS object")
                north = False

        else:

            self._data, self._header, self._wcs, self._wcsaxes_slices = \
                self._get_hdu(data, hdu, north, convention=convention,
                              dimensions=dimensions, slices=slices)
            self._wcs.nx = self._header['NAXIS%i' % (dimensions[0] + 1)]
            self._wcs.ny = self._header['NAXIS%i' % (dimensions[1] + 1)]

        # Downsample if requested
        if downsample:
            nx_new = self._wcs.nx - np.mod(self._wcs.nx, downsample)
            ny_new = self._wcs.ny - np.mod(self._wcs.ny, downsample)
            self._data = self._data[0:ny_new, 0:nx_new]
            self._data = block_reduce(self._data, downsample, func=np.mean)
            self._wcs.nx, self._wcs.ny = nx_new, ny_new

        # Open the figure
        if figure:
            self._figure = figure
        else:
            self._figure = plt.figure(**kwargs)

        # Set whether to automatically refresh the display
        self.set_auto_refresh(auto_refresh)

        # Initialize axis instance
        if type(subplot) is list and len(subplot) == 4:
            self.ax = WCSAxes(self._figure, subplot, wcs=self._wcs,
                              slices=self._wcsaxes_slices, adjustable='datalim')
            self._figure.add_axes(self.ax)
        elif type(subplot) is tuple and len(subplot) == 3:
            self.ax = WCSAxesSubplot(self._figure, *subplot, wcs=self._wcs,
                                     slices=self._wcsaxes_slices)
            self._figure.add_subplot(self.ax)
        else:
            raise ValueError("subplot= should be either a tuple of three "
                             "values, or a list of four values")

        # Turn off autoscaling
        self.ax.set_autoscale_on(False)

        # Make sure axes are above everything else
        self.ax.set_axisbelow(False)

        # Set view to whole FITS file
        self._initialize_view()

        # Set the coordinates for x and y axis
        self.x = dimensions[0]
        self.y = dimensions[1]

        # Initialize tick, label, and frame convenience wrappers (these dispatch
        # calls to WCSAxes)
        self.ticks = Ticks(self)
        self.axis_labels = AxisLabels(self)
        self.tick_labels = TickLabels(self)
        self.frame = Frame(self)

        # Display minor ticks
        self.ax.coords[self.x].display_minor_ticks(True)
        self.ax.coords[self.y].display_minor_ticks(True)

        # Initialize layers list
        self._initialize_layers()

        # Set image holder to be empty
        self.image = None

        # Set default theme
        self.set_theme(theme='pretty')

    def _get_hdu(self, data, hdu, north, convention=None,
                 dimensions=[0, 1], slices=[]):

        if isinstance(data, str):

            filename = data

            # Check file exists
            if not os.path.exists(filename):
                raise IOError("File not found: " + filename)

            # Read in FITS file
            try:
                hdulist = fits.open(filename)
            except Exception:
                raise IOError("An error occurred while reading the FITS file")

            # Check whether the HDU specified contains any data, otherwise
            # cycle through all HDUs to find one that contains valid image data
            if hdulist[hdu].data is None:
                found = False
                for alt_hdu in range(len(hdulist)):
                    if isinstance(hdulist[alt_hdu], HDU_TYPES):
                        if hdulist[alt_hdu].data is not None:
                            log.warning("hdu=%i does not contain any data, "
                                        "using hdu=%i instead" % (hdu, alt_hdu))
                            hdu = hdulist[alt_hdu]
                            found = True
                            break
                if not found:
                    raise Exception("FITS file does not contain any image data")

            else:
                hdu = hdulist[hdu]

        elif type(data) is np.ndarray:

            hdu = fits.ImageHDU(data)

        elif isinstance(data, HDU_TYPES):

            hdu = data

        elif isinstance(data, fits.HDUList):

            hdu = data[hdu]

        else:

            raise Exception("data argument should either be a filename, an HDU object from astropy.io.fits, a WCS object from astropy.wcs, or a Numpy array.")

        # Check that we have at least 2-dimensional data
        if hdu.header['NAXIS'] < 2:
            raise ValueError("Data should have at least two dimensions")

        # Check dimensions= argument
        if type(dimensions) not in [list, tuple]:
            raise ValueError('dimensions= should be a list or a tuple')
        if len(set(dimensions)) != 2 or len(dimensions) != 2:
            raise ValueError("dimensions= should be a tuple of two different values")
        if dimensions[0] < 0 or dimensions[0] > hdu.header['NAXIS'] - 1:
            raise ValueError('values of dimensions= should be between %i and %i' % (0, hdu.header['NAXIS'] - 1))
        if dimensions[1] < 0 or dimensions[1] > hdu.header['NAXIS'] - 1:
            raise ValueError('values of dimensions= should be between %i and %i' % (0, hdu.header['NAXIS'] - 1))

        # Reproject to face north if requested
        if north:

            # Find rotated WCS
            frame = ICRS() if north is True else north
            from reproject.mosaicking import find_optimal_celestial_wcs
            wcs, shape = find_optimal_celestial_wcs([hdu], frame=frame)

            from reproject import reproject_interp
            data, _ = reproject_interp(hdu, wcs, shape_out=shape)
            header = wcs.to_header()
            header['NAXIS1'] = shape[1]
            header['NAXIS2'] = shape[0]

        else:

            # Now copy the data and header to new objects, since in astropy.io.fits
            # the two attributes are linked, which can lead to confusing behavior.
            # We just need to copy the header to avoid memory issues - as long as
            # one item is copied, the two variables are decoupled.
            data = hdu.data
            header = hdu.header.copy()

        del hdu

        # If slices wasn't specified, check if we can guess
        shape = data.shape
        if len(shape) > 2:
            n_total = reduce(operator.mul, shape)
            n_image = (shape[len(shape) - 1 - dimensions[0]] *
                       shape[len(shape) - 1 - dimensions[1]])
            if n_total == n_image:
                slices = [0 for i in range(1, len(shape) - 1)]
                log.info("Setting slices=%s" % str(slices))

        # Extract slices
        data, wcsaxes_slices = slicer.slice_hypercube(data, header,
                                                      dimensions=dimensions,
                                                      slices=slices)

        # Check header
        header = header_util.check(header, convention=convention,
                                   dimensions=dimensions)

        # Parse WCS info
        wcs = WCS(header, relax=True)

        return data, header, wcs, wcsaxes_slices

    @auto_refresh
    def set_title(self, title, **kwargs):
        """
        Set the figure title
        """
        self.ax.set_title(title, **kwargs)

    @auto_refresh
    def set_xaxis_coord_type(self, coord_type):
        """
        Set the type of x coordinate.

        Options are:

        * ``scalar``: treat the values are normal decimal scalar values
        * ``longitude``: treat the values as a longitude in the 0 to 360 range
        * ``latitude``: treat the values as a latitude in the -90 to 90 range
        """
        self.ax.coords[self.x].set_coord_type(coord_type)

    @auto_refresh
    def set_yaxis_coord_type(self, coord_type):
        """
        Set the type of y coordinate.

        Options are:

        * ``scalar``: treat the values are normal decimal scalar values
        * ``longitude``: treat the values as a longitude in the 0 to 360 range
        * ``latitude``: treat the values as a latitude in the -90 to 90 range
        """
        self.ax.coords[self.y].set_coord_type(coord_type)

    @auto_refresh
    def set_system_latex(self, usetex):
        """
        Set whether to use a real LaTeX installation or the built-in matplotlib
        LaTeX.

        Parameters
        ----------

        usetex : str
            Whether to use a real LaTex installation (True) or the built-in
            matplotlib LaTeX (False). Note that if the former is chosen, an
            installation of LaTex is required.
        """
        plt.rc('text', usetex=usetex)

    @auto_refresh
    def recenter(self, x, y, radius=None, width=None, height=None):
        """
        Center the image on a given position and with a given radius.

        Either the radius or width/height arguments should be specified. The
        units of the radius or width/height should be the same as the world
        coordinates in the WCS. For images of the sky, this is often (but not
        always) degrees.

        Parameters
        ----------

        x, y : float
            Coordinates to center on

        radius : float, optional
            Radius of the region to view in degrees. This produces a square plot.

        width : float, optional
            Width of the region to view. This should be given in
            conjunction with the height argument.

        height : float, optional
            Height of the region to view. This should be given in
            conjunction with the width argument.
        """

        xpix, ypix = self.world2pixel(x, y)

        pix_scale = proj_plane_pixel_scales(self._wcs)
        sx, sy = pix_scale[self.x], pix_scale[self.y]

        if radius:
            dx_pix = radius / sx
            dy_pix = radius / sy
        elif width and height:
            dx_pix = width / sx * 0.5
            dy_pix = height / sy * 0.5
        else:
            raise Exception("Need to specify either radius= or width= and height= arguments")

        if (xpix + dx_pix < -0.5 or
            xpix - dx_pix > self._wcs.nx - 0.5 or
            ypix + dy_pix < -0.5 or
                ypix - dy_pix > self._wcs.ny):

            raise Exception("Zoom region falls outside the image")

        self.ax.set_xlim(xpix - dx_pix, xpix + dx_pix)
        self.ax.set_ylim(ypix - dy_pix, ypix + dy_pix)

    @auto_refresh
    def show_grayscale(self, vmin=None, vmid=None, vmax=None,
                       pmin=0.25, pmax=99.75,
                       stretch='linear', exponent=2, invert='default',
                       smooth=None, kernel='gauss', aspect='equal',
                       interpolation='nearest'):
        """
        Show a grayscale image of the FITS file.

        Parameters
        ----------

        vmin : None or float, optional
            Minimum pixel value to use for the grayscale. If set to None,
            the minimum pixel value is determined using pmin (default).

        vmax : None or float, optional
            Maximum pixel value to use for the grayscale. If set to None,
            the maximum pixel value is determined using pmax (default).

        pmin : float, optional
            Percentile value used to determine the minimum pixel value to
            use for the grayscale if vmin is set to None. The default
            value is 0.25%.

        pmax : float, optional
            Percentile value used to determine the maximum pixel value to
            use for the grayscale if vmax is set to None. The default
            value is 99.75%.

        stretch : { 'linear', 'log', 'sqrt', 'arcsinh', 'power' }, optional
            The stretch function to use

        vmid : None or float, optional
            Baseline value used for the log and arcsinh stretches. If
            set to None, this is set to zero for log stretches and to
            vmin - (vmax - vmin) / 30. for arcsinh stretches

        exponent : float, optional
            If stretch is set to 'power', this is the exponent to use

        invert : str, optional
            Whether to invert the grayscale or not. The default is False,
            unless set_theme is used, in which case the default depends on
            the theme.

        smooth : int or tuple, optional
            Default smoothing scale is 3 pixels across. User can define
            whether they want an NxN kernel (integer), or NxM kernel
            (tuple). This argument corresponds to the 'gauss' and 'box'
            smoothing kernels.

        kernel : { 'gauss', 'box', numpy.array }, optional
            Default kernel used for smoothing is 'gauss'. The user can
            specify if they would prefer 'gauss', 'box', or a custom
            kernel. All kernels are normalized to ensure flux retention.

        aspect : { 'auto', 'equal' }, optional
            Whether to change the aspect ratio of the image to match that
            of the axes ('auto') or to change the aspect ratio of the axes
            to match that of the data ('equal'; default)

        interpolation : str, optional
            The type of interpolation to use for the image. The default is
            'nearest'. Other options include 'none' (no interpolation,
            meaning that if exported to a postscript file, the grayscale
            will be output at native resolution irrespective of the dpi
            setting), 'bilinear', 'bicubic', and many more (see the
            matplotlib documentation for imshow).
        """

        if invert == 'default':
            invert = self._get_invert_default()

        if invert:
            cmap = 'gist_yarg'
        else:
            cmap = 'gray'

        self.show_colorscale(vmin=vmin, vmid=vmid, vmax=vmax,
                             pmin=pmin, pmax=pmax,
                             stretch=stretch, exponent=exponent, cmap=cmap,
                             smooth=smooth, kernel=kernel, aspect=aspect,
                             interpolation=interpolation)

    @auto_refresh
    def hide_grayscale(self, *args, **kwargs):
        self.hide_colorscale(*args, **kwargs)

    @auto_refresh
    def show_colorscale(self, vmin=None, vmid=None, vmax=None, pmin=0.25,
                        pmax=99.75, stretch='linear', exponent=2,
                        cmap='default', smooth=None, kernel='gauss',
                        aspect='equal', interpolation='nearest'):
        """
        Show a colorscale image of the FITS file.

        Parameters
        ----------

        vmin : None or float, optional
            Minimum pixel value to use for the colorscale. If set to None,
            the minimum pixel value is determined using pmin (default).

        vmax : None or float, optional
            Maximum pixel value to use for the colorscale. If set to None,
            the maximum pixel value is determined using pmax (default).

        pmin : float, optional
            Percentile value used to determine the minimum pixel value to
            use for the colorscale if vmin is set to None. The default
            value is 0.25%.

        pmax : float, optional
            Percentile value used to determine the maximum pixel value to
            use for the colorscale if vmax is set to None. The default
            value is 99.75%.

        stretch : { 'linear', 'log', 'sqrt', 'arcsinh', 'power' }, optional
            The stretch function to use

        vmid : None or float, optional
            Baseline value used for the log and arcsinh stretches. If
            not set, this defaults to zero for log stretches and to
            vmin - (vmax - vmin) / 30. for arcsinh stretches

        exponent : float, optional
            If stretch is set to 'power', this is the exponent to use

        cmap : str, optional
            The name of the colormap to use

        smooth : int or tuple, optional
            Default smoothing scale is 3 pixels across. User can define
            whether they want an NxN kernel (integer), or NxM kernel
            (tuple). This argument corresponds to the 'gauss' and 'box'
            smoothing kernels.

        kernel : { 'gauss', 'box', numpy.array }, optional
            Default kernel used for smoothing is 'gauss'. The user can
            specify if they would prefer 'gauss', 'box', or a custom
            kernel. All kernels are normalized to ensure flux retention.

        aspect : { 'auto', 'equal' }, optional
            Whether to change the aspect ratio of the image to match that
            of the axes ('auto') or to change the aspect ratio of the axes
            to match that of the data ('equal'; default)

        interpolation : str, optional
            The type of interpolation to use for the image. The default is
            'nearest'. Other options include 'none' (no interpolation,
            meaning that if exported to a postscript file, the colorscale
            will be output at native resolution irrespective of the dpi
            setting), 'bilinear', 'bicubic', and many more (see the
            matplotlib documentation for imshow).
        """

        if cmap == 'default':
            cmap = self._get_colormap_default()

        min_auto = vmin is None
        max_auto = vmax is None

        # The set of available functions
        cmap = plt.get_cmap(cmap)

        if min_auto or max_auto:

            interval = AsymmetricPercentileInterval(pmin, pmax, n_samples=10000)
            try:
                vmin_auto, vmax_auto = interval.get_limits(self._data)
            except (IndexError, TypeError):  # no valid values
                vmin_auto = vmax_auto = 0

            if min_auto:
                vmin = vmin_auto

            if max_auto:
                vmax = vmax_auto

        # Prepare normalizer object
        if stretch == 'arcsinh':
            stretch = 'asinh'

        if stretch == 'log':
            if vmid is None:
                if vmin < 0:
                    raise ValueError("When using a log stretch, if vmin < 0, then vmid has to be specified")
                else:
                    vmid = 0.
            if vmin < vmid:
                raise ValueError("When using a log stretch, vmin should be larger than vmid")
            log_a = (vmax - vmid) / (vmin - vmid)
            norm_kwargs = {'log_a': log_a}
        elif stretch == 'asinh':
            if vmid is None:
                vmid = vmin - (vmax - vmin) / 30.
            asinh_a = (vmid - vmin) / (vmax - vmin)
            norm_kwargs = {'asinh_a': abs(asinh_a)}
        else:
            norm_kwargs = {}

        normalizer = simple_norm(self._data, stretch=stretch, power=exponent,
                                 min_cut=vmin, max_cut=vmax, clip=False,
                                 **norm_kwargs)

        # Adjust vmin/vmax if auto
        if min_auto:
            if stretch == 'linear':
                vmin = -0.1 * (vmax - vmin) + vmin
            log.info("Auto-setting vmin to %10.3e" % vmin)

        if max_auto:
            if stretch == 'linear':
                vmax = 0.1 * (vmax - vmin) + vmax
            log.info("Auto-setting vmax to %10.3e" % vmax)

        # Update normalizer object
        normalizer.vmin = vmin
        normalizer.vmax = vmax

        if self.image:
            self.image.set_visible(True)
            self.image.set_norm(normalizer)
            self.image.set_cmap(cmap=cmap)
            self.image.origin = 'lower'
            self.image.set_interpolation(interpolation)
            self.image.set_data(convolve_util.convolve(self._data,
                                                       smooth=smooth,
                                                       kernel=kernel))
        else:
            extent = -0.5, self._wcs.nx - 0.5, -0.5, self._wcs.ny - 0.5
            convolved_data = convolve_util.convolve(self._data, smooth=smooth, kernel=kernel)
            self.image = self.ax.imshow(convolved_data, cmap=cmap,
                                        interpolation=interpolation,
                                        origin='lower', norm=normalizer,
                                        aspect=aspect, extent=extent)

        xmin, xmax = self.ax.get_xbound()
        if xmin == 0.0:
            self.ax.set_xlim(0.5, xmax)

        ymin, ymax = self.ax.get_ybound()
        if ymin == 0.0:
            self.ax.set_ylim(0.5, ymax)

        if hasattr(self, 'colorbar'):
            self.colorbar.update()

    @auto_refresh
    def hide_colorscale(self):
        self.image.set_visible(False)

    @auto_refresh
    def set_nan_color(self, color):
        """
        Set the color for NaN pixels.

        Parameters
        ----------
        color : str
            This can be any valid matplotlib color
        """
        from copy import deepcopy
        cm = deepcopy(self.image.get_cmap())
        cm.set_bad(color)
        self.image.set_cmap(cm)

    @auto_refresh
    def show_rgb(self, filename=None, interpolation='nearest',
                 vertical_flip=False, horizontal_flip=False, flip=False):
        """
        Show a 3-color image instead of the FITS file data.

        Parameters
        ----------

        filename, optional
            The 3-color image should have exactly the same dimensions
            as the FITS file, and will be shown with exactly the same
            projection. If FITSFigure was initialized with an
            AVM-tagged RGB image, the filename is not needed here.

        vertical_flip : str, optional
            Whether to vertically flip the RGB image

        horizontal_flip : str, optional
            Whether to horizontally flip the RGB image
        """

        try:
            from PIL import Image
        except ImportError:
            try:
                import Image
            except ImportError:
                raise ImportError("The Python Imaging Library (PIL) is required to read in RGB images")

        if flip:
            log.warning("Note that show_rgb should now correctly flip RGB images, so the flip= argument is now deprecated. If you still need to flip an image vertically or horizontally, you can use the vertical_flip= and horizontal_flip arguments instead.")

        if filename is None:
            if hasattr(self, '_rgb_image'):
                image = Image.open(self._rgb_image)
            else:
                raise Exception("Need to specify the filename of an RGB image")
        else:
            image = Image.open(filename)

        if vertical_flip:
            image = image.transpose(Image.FLIP_TOP_BOTTOM)

        if horizontal_flip:
            image = image.transpose(Image.FLIP_LEFT_RIGHT)

        self.image = self.ax.imshow(image,
                                    interpolation=interpolation,
                                    origin='lower')

    @auto_refresh
    def show_contour(self, data=None, hdu=0, layer=None, levels=5,
                     filled=False, cmap=None, colors=None, returnlevels=False,
                     convention=None, dimensions=[0, 1], slices=[],
                     smooth=None, kernel='gauss', overlap=False, **kwargs):
        """
        Overlay contours on the current plot.

        Parameters
        ----------

        data : see below

            The FITS file to plot contours for. The following data types can be
            passed:

                 string
                 astropy.io.fits.PrimaryHDU
                 astropy.io.fits.ImageHDU
                 astropy.wcs.WCS
                 np.ndarray

        hdu : int, optional
            By default, the image in the primary HDU is read in. If a
            different HDU is required, use this argument.

        layer : str, optional
            The name of the contour layer. This is useful for giving
            custom names to layers (instead of contour_set_n) and for
            replacing existing layers.

        levels : int or list, optional
            This can either be the number of contour levels to compute
            (if an integer is provided) or the actual list of contours
            to show (if a list of floats is provided)

        filled : str, optional
            Whether to show filled or line contours

        cmap : str, optional
            The colormap to use for the contours

        colors : str or tuple, optional
            If a single string is provided, all contour levels will be
            shown in this color. If a tuple of strings is provided,
            each contour will be colored according to the corresponding
            tuple element.

        returnlevels : str, optional
            Whether to return the list of contours to the caller.

        convention : str, optional
            This is used in cases where a FITS header can be interpreted
            in multiple ways. For example, for files with a -CAR
            projection and CRVAL2=0, this can be set to 'wells' or
            'calabretta' to choose the appropriate convention.

        dimensions : tuple or list, optional
            The index of the axes to use if the data has more than three
            dimensions.

        slices : tuple or list, optional
            If a FITS file with more than two dimensions is specified,
            then these are the slices to extract. If all extra dimensions
            only have size 1, then this is not required.

        smooth : int or tuple, optional
            Default smoothing scale is 3 pixels across. User can define
            whether they want an NxN kernel (integer), or NxM kernel
            (tuple). This argument corresponds to the 'gauss' and 'box'
            smoothing kernels.

        kernel : { 'gauss' , 'box' , numpy.array }, optional
            Default kernel used for smoothing is 'gauss'. The user can
            specify if they would prefer 'gauss', 'box', or a custom
            kernel. All kernels are normalized to ensure flux retention.

        overlap str, optional
            Whether to include only contours that overlap with the image
            area. This significantly speeds up the drawing of contours and
            reduces file size when using a file for the contours covering
            a much larger area than the image.

        kwargs
            Additional keyword arguments (such as alpha, linewidths, or
            linestyles) will be passed on directly to Matplotlib's
            :meth:`~matplotlib.axes.Axes.contour` or
            :meth:`~matplotlib.axes.Axes.contourf` methods. For more
            information on these additional arguments, see the *Optional
            keyword arguments* sections in the documentation for those
            methods.
        """
        if layer:
            self.remove_layer(layer, raise_exception=False)

        if cmap:
            cmap = plt.get_cmap(cmap)
        elif not colors:
            cmap = plt.get_cmap('viridis')

        if data is not None:
            data_contour, header_contour, wcs_contour, wcsaxes_slices = \
                self._get_hdu(data, hdu, False, convention=convention,
                              dimensions=dimensions, slices=slices)
        else:
            data_contour = self._data
            header_contour = self._header
            wcs_contour = self._wcs

        wcs_contour.nx = header_contour['NAXIS%i' % (dimensions[0] + 1)]
        wcs_contour.ny = header_contour['NAXIS%i' % (dimensions[1] + 1)]

        image_contour = convolve_util.convolve(data_contour, smooth=smooth, kernel=kernel)
        if type(levels) is int:
            interval = AsymmetricPercentileInterval(0.25, 99.75, n_samples=10000)
            try:
                vmin_auto, vmax_auto = interval.get_limits(image_contour)
            except IndexError:  # no valid values
                vmin_auto = vmax_auto = 0
            levels = np.linspace(vmin_auto, vmax_auto, levels)

        if wcs_contour.wcs.ctype[self.x] == 'PIXEL' or wcs_contour.wcs.ctype[self.y] == 'PIXEL':
            frame = 'pixel'
        else:
            frame = wcs_contour

        if filled:
            c = self.ax.contourf(image_contour, levels,
                                 transform=self.ax.get_transform(frame),
                                 cmap=cmap,
                                 colors=colors, **kwargs)
        else:
            c = self.ax.contour(image_contour, levels,
                                transform=self.ax.get_transform(frame),
                                cmap=cmap,
                                colors=colors, **kwargs)

        if layer:
            contour_set_name = layer
        else:
            self._contour_counter += 1
            contour_set_name = 'contour_set_' + str(self._contour_counter)

        self._layers[contour_set_name] = c

        if returnlevels:
            return levels

    @auto_refresh
    def show_vectors(self, pdata, adata, phdu=0, ahdu=0, step=1, scale=1,
                     rotate=0, cutoff=0, units='degrees', layer=None,
                     convention=None, dimensions=[0, 1], slices=[], **kwargs):
        """
        Overlay vectors on the current plot.

        Parameters
        ----------

        pdata : see below

            The FITS file specifying the magnitude of vectors. The following data types can be passed:

                 string
                 astropy.io.fits.PrimaryHDU
                 astropy.io.fits.ImageHDU
                 astropy.wcs.WCS
                 np.ndarray

        adata : see below

            The FITS file specifying the angle of vectors. The following data types can be passed:

                 string
                 astropy.io.fits.PrimaryHDU
                 astropy.io.fits.ImageHDU
                 astropy.wcs.WCS
                 np.ndarray

        phdu : int, optional
            By default, the image in the primary HDU is read in. If a
            different HDU is required for pdata, use this argument.

        ahdu : int, optional
            By default, the image in the primary HDU is read in. If a
            different HDU is required for adata, use this argument.

        step : int, optional
            Derive a vector only from every 'step' pixels. You will
            normally want this to be >1 to get sensible vector spacing.

        scale : int, optional
            The length, in pixels, of a vector with magnitude 1 in the image
            specified by pdata. If pdata specifies fractional polarization,
            make this comparable to step.

        rotate : float, optional
            An angle to rotate by, in units the same as those of the angle map.

        cutoff : float, optional
            The value of magnitude below which no
            vectors should be plotted. The default value, zero,
            excludes negative-length and NaN-masked data.

        units : str, optional
            Units to assume for the angle map. Valid values are 'degrees'
            (the default) or 'radians' (or anything else), which will not
            apply a scaling factor of pi/180 to the angle data.

        layer : str, optional
            The name of the vector layer. This is useful for giving
            custom names to layers (instead of vector_set_n) and for
            replacing existing layers.

        convention : str, optional
            This is used in cases where a FITS header can be interpreted
            in multiple ways. For example, for files with a -CAR
            projection and CRVAL2=0, this can be set to 'wells' or
            'calabretta' to choose the appropriate convention.

        dimensions : tuple or list, optional
            The index of the axes to use if the data has more than three
            dimensions.

        slices : tuple or list, optional
            If a FITS file with more than two dimensions is specified,
            then these are the slices to extract. If all extra dimensions
            only have size 1, then this is not required.

        kwargs
            Additional keyword arguments (such as alpha, linewidths, or color)
            which are passed to Matplotlib's
            :class:`~matplotlib.collections.LineCollection` class, and can be used to
            control the appearance of the lines. For more
            information on these additional arguments, see the *Optional
            keyword arguments* sections in the documentation for those
            methods.

        """

        # over-ride default color (none) that will otherwise be set by
        # show_lines()
        if 'color' not in kwargs:
            kwargs.setdefault('color', 'black')

        if layer:
            self.remove_layer(layer, raise_exception=False)

        data_p, header_p, wcs_p, slices_p = \
            self._get_hdu(pdata, phdu, False, convention=convention,
                          dimensions=dimensions, slices=slices)
        data_a, header_a, wcs_a, slices_a = \
            self._get_hdu(adata, ahdu, False, convention=convention,
                          dimensions=dimensions, slices=slices)

        # TODO: use slices correctly

        wcs_p.nx = header_p['NAXIS%i' % (dimensions[0] + 1)]
        wcs_p.ny = header_p['NAXIS%i' % (dimensions[1] + 1)]
        wcs_a.nx = header_a['NAXIS%i' % (dimensions[0] + 1)]
        wcs_a.ny = header_a['NAXIS%i' % (dimensions[1] + 1)]

        if (wcs_p.nx != wcs_a.nx or wcs_p.ny != wcs_a.ny):
            raise Exception("Angle and magnitude images must be same size")

        angle = data_a + rotate
        if units == 'degrees':
            angle = np.radians(angle)

        linelist = []
        for y in range(0, wcs_p.ny, step):
            for x in range(0, wcs_p.nx, step):
                if data_p[y, x] > cutoff and np.isfinite(angle[y, x]):
                    r = data_p[y, x] * 0.5 * scale
                    a = angle[y, x]
                    x1 = x + r * np.sin(a)
                    y1 = y - r * np.cos(a)
                    x2 = x - r * np.sin(a)
                    y2 = y + r * np.cos(a)

                    x_world, y_world = self.pixel2world([x1, x2], [y1, y2], wcs=wcs_p)
                    line = np.array([x_world, y_world])
                    linelist.append(line)

        if layer:
            vector_set_name = layer
        else:
            self._vector_counter += 1
            vector_set_name = 'vector_set_' + str(self._vector_counter)

        # Use show_lines to finish the process off
        self.show_lines(linelist, layer=vector_set_name, **kwargs)

    # This method plots markers. The input should be an Nx2 array with WCS
    # coordinates in degree format.

    @auto_refresh
    def show_markers(self, xw, yw, layer=False, coords_frame='world', **kwargs):
        """
        Overlay markers on the current plot.

        Parameters
        ----------

        xw : list or `~numpy.ndarray`
            The x positions of the markers (in world coordinates)

        yw : list or `~numpy.ndarray`
            The y positions of the markers (in world coordinates)

        layer : str, optional
            The name of the scatter layer. This is useful for giving
            custom names to layers (instead of marker_set_n) and for
            replacing existing layers.

        coords_frame : 'pixel' or 'world'
            The reference frame in which the coordinates are defined. This is
            used to interpret the values of ``xw`` and ``yw``.

        kwargs
            Additional keyword arguments (such as marker, facecolor,
            edgecolor, alpha, or linewidth) will be passed on directly to
            Matplotlib's :meth:`~matplotlib.axes.Axes.scatter` method (in
            particular, have a look at the *Optional keyword arguments* in the
            documentation for that method).
        """

        if 'c' not in kwargs:
            kwargs.setdefault('edgecolor', 'red')
            kwargs.setdefault('facecolor', 'none')

        kwargs.setdefault('s', 30)

        if layer:
            self.remove_layer(layer, raise_exception=False)

        s = self.ax.scatter(xw, yw, transform=self.ax.get_transform(coords_frame), **kwargs)

        if layer:
            marker_set_name = layer
        else:
            self._scatter_counter += 1
            marker_set_name = 'marker_set_' + str(self._scatter_counter)

        self._layers[marker_set_name] = s

    # Show circles. Different from markers as this method allows more
    # definitions for the circles.
    @auto_refresh
    def show_circles(self, xw, yw, radius, layer=False, coords_frame='world', zorder=None, **kwargs):
        """
        Overlay circles on the current plot.

        Parameters
        ----------

        xw : list or `~numpy.ndarray`
            The x positions of the centers of the circles (in world coordinates)

        yw : list or `~numpy.ndarray`
            The y positions of the centers of the circles (in world coordinates)

        radius : int or float or list or `~numpy.ndarray`
            The radii of the circles (in world coordinates)

        layer : str, optional
            The name of the circle layer. This is useful for giving
            custom names to layers (instead of circle_set_n) and for
            replacing existing layers.

        coords_frame : 'pixel' or 'world'
            The reference frame in which the coordinates are defined. This is
            used to interpret the values of ``xw`` and ``yw``.

        kwargs
            Additional keyword arguments (such as facecolor, edgecolor, alpha,
            or linewidth) are passed to Matplotlib
            :class:`~matplotlib.collections.PatchCollection` class, and can be
            used to control the appearance of the circles.
        """

        xw, yw, radius = uniformize_1d(xw, yw, radius)

        if 'facecolor' not in kwargs:
            kwargs.setdefault('facecolor', 'none')

        if layer:
            self.remove_layer(layer, raise_exception=False)

        if coords_frame not in ['pixel', 'world']:
            raise ValueError("coords_frame should be set to 'pixel' or 'world'")

        # While we could plot the shape using the get_transform('world') mode
        # from WCSAxes, the issue is that the rotation angle is also measured in
        # world coordinates so will not be what the user is expecting. So we allow the user to specify the reference frame for the coordinates and for the rotation.

        if coords_frame == 'pixel':
            x, y = xw, yw
            r = radius
        else:
            x, y = self.world2pixel(xw, yw)
            pix_scale = proj_plane_pixel_scales(self._wcs)
            sx, sy = pix_scale[self.x], pix_scale[self.y]
            r = radius / np.sqrt(sx * sy)

        patches = []
        for i in range(len(xw)):
            patches.append(Circle((x[i], y[i]), radius=r[i]))

        # Due to bugs in matplotlib, we need to pass the patch properties
        # directly to the PatchCollection rather than use match_original.
        p = PatchCollection(patches, **kwargs)

        if zorder is not None:
            p.zorder = zorder
        c = self.ax.add_collection(p)

        if layer:
            circle_set_name = layer
        else:
            self._circle_counter += 1
            circle_set_name = 'circle_set_' + str(self._circle_counter)

        self._layers[circle_set_name] = c

    @auto_refresh
    def show_ellipses(self, xw, yw, width, height, angle=0, layer=False,
                      zorder=None, coords_frame='world', **kwargs):
        """
        Overlay ellipses on the current plot.

        Parameters
        ----------

        xw : list or `~numpy.ndarray`
            The x positions of the centers of the ellipses (in world coordinates)

        yw : list or `~numpy.ndarray`
            The y positions of the centers of the ellipses (in world coordinates)

        width : int or float or list or `~numpy.ndarray`
            The width of the ellipse (in world coordinates)

        height : int or float or list or `~numpy.ndarray`
            The height of the ellipse (in world coordinates)

        angle : int or float or list or `~numpy.ndarray`, optional
            rotation in degrees (anti-clockwise). Default
            angle is 0.0.

        layer : str, optional
            The name of the ellipse layer. This is useful for giving
            custom names to layers (instead of ellipse_set_n) and for
            replacing existing layers.

        coords_frame : 'pixel' or 'world'
            The reference frame in which the coordinates are defined. This is
            used to interpret the values of ``xw``, ``yw``, ``width``, and
            ``height``.

        kwargs
            Additional keyword arguments (such as facecolor, edgecolor, alpha,
            or linewidth) are passed to Matplotlib
            :class:`~matplotlib.collections.PatchCollection` class, and can be
            used to control the appearance of the ellipses.
        """

        xw, yw, width, height, angle = uniformize_1d(xw, yw, width, height, angle)

        if 'facecolor' not in kwargs:
            kwargs.setdefault('facecolor', 'none')

        if 'edgecolor' not in kwargs:
            kwargs.setdefault('edgecolor', 'black')

        if layer:
            self.remove_layer(layer, raise_exception=False)

        if coords_frame not in ['pixel', 'world']:
            raise ValueError("coords_frame should be set to 'pixel' or 'world'")

        # While we could plot the shape using the get_transform('world') mode
        # from WCSAxes, the issue is that the rotation angle is also measured in
        # world coordinates so will not be what the user is expecting. So we allow the user to specify the reference frame for the coordinates and for the rotation.

        if coords_frame == 'pixel':
            x, y = xw, yw
            w = width
            h = height
            a = angle
            transform = self.ax.transData
        else:
            x, y = self.world2pixel(xw, yw)
            pix_scale = proj_plane_pixel_scales(self._wcs)
            sx, sy = pix_scale[self.x], pix_scale[self.y]
            w = width / sx
            h = height / sy
            a = angle
            transform = self.ax.transData

        patches = []
        for i in range(len(x)):
            patches.append(Ellipse((x[i], y[i]), width=w[i], height=h[i], angle=a[i]))

        # Due to bugs in matplotlib, we need to pass the patch properties
        # directly to the PatchCollection rather than use match_original.
        p = PatchCollection(patches, transform=transform, **kwargs)

        if zorder is not None:
            p.zorder = zorder
        c = self.ax.add_collection(p)

        if layer:
            ellipse_set_name = layer
        else:
            self._ellipse_counter += 1
            ellipse_set_name = 'ellipse_set_' + str(self._ellipse_counter)

        self._layers[ellipse_set_name] = c

    @auto_refresh
    def show_rectangles(self, xw, yw, width, height, angle=0, layer=False,
                        zorder=None, coords_frame='world', **kwargs):
        """
        Overlay rectangles on the current plot.

        Parameters
        ----------

        xw : list or `~numpy.ndarray`
            The x positions of the centers of the rectangles (in world coordinates)

        yw : list or `~numpy.ndarray`
            The y positions of the centers of the rectangles (in world coordinates)

        width : int or float or list or `~numpy.ndarray`
            The width of the rectangle (in world coordinates)

        height : int or float or list or `~numpy.ndarray`
            The height of the rectangle (in world coordinates)

        angle : int or float or list or `~numpy.ndarray`, optional
            rotation in degrees (anti-clockwise). Default
            angle is 0.0.

        layer : str, optional
            The name of the rectangle layer. This is useful for giving
            custom names to layers (instead of rectangle_set_n) and for
            replacing existing layers.

        coords_frame : 'pixel' or 'world'
            The reference frame in which the coordinates are defined. This is
            used to interpret the values of ``xw``, ``yw``, ``width``, and
            ``height``.

        kwargs
            Additional keyword arguments (such as facecolor, edgecolor, alpha,
            or linewidth) are passed to Matplotlib
            :class:`~matplotlib.collections.PatchCollection` class, and can be
            used to control the appearance of the rectangles.
        """

        xw, yw, width, height, angle = uniformize_1d(xw, yw, width, height, angle)

        if 'facecolor' not in kwargs:
            kwargs.setdefault('facecolor', 'none')

        if layer:
            self.remove_layer(layer, raise_exception=False)

        if coords_frame not in ['pixel', 'world']:
            raise ValueError("coords_frame should be set to 'pixel' or 'world'")

        # While we could plot the shape using the get_transform('world') mode
        # from WCSAxes, the issue is that the rotation angle is also measured in
        # world coordinates so will not be what the user is expecting. So we
        # allow the user to specify the reference frame for the coordinates and
        # for the rotation.

        if coords_frame == 'pixel':
            x, y = xw, yw
            w = width
            h = height
            a = angle
            transform = self.ax.transData
        else:
            x, y = self.world2pixel(xw, yw)
            pix_scale = proj_plane_pixel_scales(self._wcs)
            sx, sy = pix_scale[self.x], pix_scale[self.y]
            w = width / sx
            h = height / sy
            a = angle
            transform = self.ax.transData

        x = x - w / 2.
        y = y - h / 2.

        patches = []
        for i in range(len(x)):
            patches.append(Rectangle((x[i], y[i]), width=w[i], height=h[i], angle=a[i]))

        # Due to bugs in matplotlib, we need to pass the patch properties
        # directly to the PatchCollection rather than use match_original.
        p = PatchCollection(patches, transform=transform, **kwargs)

        if zorder is not None:
            p.zorder = zorder
        c = self.ax.add_collection(p)

        if layer:
            rectangle_set_name = layer
        else:
            self._rectangle_counter += 1
            rectangle_set_name = 'rectangle_set_' + str(self._rectangle_counter)

        self._layers[rectangle_set_name] = c

    @auto_refresh
    def show_lines(self, line_list, layer=False, zorder=None, **kwargs):
        """
        Overlay lines on the current plot.

        Parameters
        ----------

        line_list : list
             A list of one or more 2xN numpy arrays which contain
             the [x, y] positions of the vertices in world coordinates.

        layer : str, optional
            The name of the line(s) layer. This is useful for giving
            custom names to layers (instead of line_set_n) and for
            replacing existing layers.

        kwargs
            Additional keyword arguments (such as color, offsets, linestyle,
            or linewidth) are passed to Matplotlib
            :class:`~matplotlib.collections.LineCollection` class, and can be used to
            control the appearance of the lines.
        """

        if 'color' not in kwargs:
            kwargs.setdefault('color', 'none')

        if layer:
            self.remove_layer(layer, raise_exception=False)

        lines = []

        for line in line_list:
            xw, yw = line[0, :], line[1, :]
            lines.append(np.column_stack((xw, yw)))

        lc = LineCollection(lines, transform=self.ax.get_transform('world'), **kwargs)
        if zorder is not None:
            lc.zorder = zorder
        c = self.ax.add_collection(lc)

        if layer:
            line_set_name = layer
        else:
            self._linelist_counter += 1
            line_set_name = 'line_set_' + str(self._linelist_counter)

        self._layers[line_set_name] = c

    @auto_refresh
    def show_arrows(self, x, y, dx, dy, width='auto', head_width='auto',
                    head_length='auto', length_includes_head=True, layer=False,
                    zorder=None, **kwargs):
        """
        Overlay arrows on the current plot.

        Parameters
        ----------

        x, y, dx, dy : float or list or `~numpy.ndarray`
            Origin and displacement of the arrows in world coordinates.
            These can either be scalars to plot a single arrow, or lists or
            arrays to plot multiple arrows.

        width : float, optional
            The width of the arrow body, in pixels (default: 2% of the
            arrow length)

        head_width : float, optional
            The width of the arrow head, in pixels (default: 5% of the
            arrow length)

        head_length : float, optional
            The length of the arrow head, in pixels (default: 5% of the
            arrow length)

        length_includes_head : bool, optional
            Whether the head includes the length

        layer : str, optional
            The name of the arrow(s) layer. This is useful for giving
            custom names to layers (instead of line_set_n) and for
            replacing existing layers.

        kwargs
            Additional keyword arguments (such as facecolor, edgecolor, alpha,
            or linewidth) are passed to Matplotlib
            :class:`~matplotlib.collections.PatchCollection` class, and can be
            used to control the appearance of the arrows.
        """

        x, y, dx, dy = uniformize_1d(x, y, dx, dy)

        if layer:
            self.remove_layer(layer, raise_exception=False)

        arrows = []

        # Here we don't make use of WCSAxes.get_transform('world') because
        # otherwise the arrow heads will be distored. Instead, we work in pixel
        # coordinates.

        for i in range(len(x)):

            xp1, yp1 = self.world2pixel(x[i], y[i])
            xp2, yp2 = self.world2pixel(x[i] + dx[i], y[i] + dy[i])

            if width == 'auto':
                width = 0.02 * np.sqrt((xp2 - xp1) ** 2 + (yp2 - yp1) ** 2)

            if head_width == 'auto':
                head_width = 0.1 * np.sqrt((xp2 - xp1) ** 2 + (yp2 - yp1) ** 2)

            if head_length == 'auto':
                head_length = 0.1 * np.sqrt((xp2 - xp1) ** 2 + (yp2 - yp1) ** 2)

            arrows.append(FancyArrow(xp1, yp1, xp2 - xp1, yp2 - yp1,
                                     width=width, head_width=head_width,
                                     head_length=head_length,
                                     length_includes_head=length_includes_head)
                          )

        # Due to bugs in matplotlib, we need to pass the patch properties
        # directly to the PatchCollection rather than use match_original.
        p = PatchCollection(arrows, **kwargs)

        if zorder is not None:
            p.zorder = zorder
        c = self.ax.add_collection(p)

        if layer:
            line_set_name = layer
        else:
            self._linelist_counter += 1
            line_set_name = 'arrow_set_' + str(self._linelist_counter)

        self._layers[line_set_name] = c

    @auto_refresh
    def show_polygons(self, polygon_list, layer=False, zorder=None, **kwargs):
        """
        Overlay polygons on the current plot.

        Parameters
        ----------

        polygon_list : list or tuple
            A list of one or more 2xN or Nx2 Numpy arrays which contain
            the [x, y] positions of the vertices in world coordinates.
            Note that N should be greater than 2.

        layer : str, optional
            The name of the circle layer. This is useful for giving
            custom names to layers (instead of circle_set_n) and for
            replacing existing layers.

        kwargs
            Additional keyword arguments (such as facecolor, edgecolor, alpha,
            or linewidth) are passed to Matplotlib
            :class:`~matplotlib.collections.PatchCollection` class, and can be
            used to control the appearance of the polygons.
        """

        if 'facecolor' not in kwargs:
            kwargs.setdefault('facecolor', 'none')

        if layer:
            self.remove_layer(layer, raise_exception=False)

        if type(polygon_list) not in [list, tuple]:
            raise Exception("polygon_list should be a list or tuple of Numpy arrays")

        pix_polygon_list = []
        for polygon in polygon_list:

            if type(polygon) is not np.ndarray:
                raise Exception("Polygon should be given as a Numpy array")

            if polygon.shape[0] == 2 and polygon.shape[1] > 2:
                xw = polygon[0, :]
                yw = polygon[1, :]
            elif polygon.shape[0] > 2 and polygon.shape[1] == 2:
                xw = polygon[:, 0]
                yw = polygon[:, 1]
            else:
                raise Exception("Polygon should have dimensions 2xN or Nx2 with N>2")

            pix_polygon_list.append(np.column_stack((xw, yw)))

        patches = []
        for i in range(len(pix_polygon_list)):
            patches.append(Polygon(pix_polygon_list[i], **kwargs))

        # Due to bugs in matplotlib, we need to pass the patch properties
        # directly to the PatchCollection rather than use match_original.
        p = PatchCollection(patches, transform=self.ax.get_transform('world'), **kwargs)

        if zorder is not None:
            p.zorder = zorder
        c = self.ax.add_collection(p)

        if layer:
            poly_set_name = layer
        else:
            self._poly_counter += 1
            poly_set_name = 'poly_set_' + str(self._poly_counter)

        self._layers[poly_set_name] = c

    @auto_refresh
    @fixdocstring
    def add_label(self, x, y, text, relative=False, color='black',
                  family=None, style=None, variant=None, stretch=None,
                  weight=None, size=None, fontproperties=None,
                  horizontalalignment='center', verticalalignment='center',
                  layer=None, **kwargs):
        """
        Add a text label.

        Parameters
        ----------

        x, y : float
            Coordinates of the text label

        text : str
            The label

        relative : str, optional
            Whether the coordinates are to be interpreted as world
            coordinates (e.g. RA/Dec or longitude/latitude), or
            coordinates relative to the axes (where 0.0 is left or bottom
            and 1.0 is right or top).

        common: color, family, style, variant, stretch, weight, size, fontproperties, horizontalalignment, verticalalignment
        """

        if layer:
            self.remove_layer(layer, raise_exception=False)

        # Can't pass fontproperties=None to text. Only pass it if it is not None.
        if fontproperties:
            kwargs['fontproperties'] = fontproperties

        if not np.isscalar(x):
            raise Exception("x should be a single value")

        if not np.isscalar(y):
            raise Exception("y should be a single value")

        if not np.isscalar(text):
            raise Exception("text should be a single value")

        if relative:
            lc = self.ax.text(x, y, text, color=color,
                              family=family, style=style, variant=variant,
                              stretch=stretch, weight=weight, size=size,
                              horizontalalignment=horizontalalignment,
                              verticalalignment=verticalalignment,
                              transform=self.ax.transAxes, **kwargs)
        else:
            lc = self.ax.text(x, y, text, color=color,
                              family=family, style=style, variant=variant,
                              stretch=stretch, weight=weight, size=size,
                              horizontalalignment=horizontalalignment,
                              verticalalignment=verticalalignment,
                              transform=self.ax.get_transform('world'),
                              **kwargs)

        if layer:
            label_name = layer
        else:
            self._label_counter += 1
            label_name = 'label_' + str(self._label_counter)

        self._layers[label_name] = lc

    def set_auto_refresh(self, refresh):
        """
        Set whether the display should refresh after each method call.

        Parameters
        ----------
        refresh : bool
            Whether to refresh the display every time a FITSFigure method is
            called. This defaults to `True` if and only if APLpy is being used
            from IPython and the Matplotlib backend is interactive.
        """
        if refresh is None:
            if matplotlib.is_interactive():
                try:
                    get_ipython()
                except NameError:
                    refresh = False
                else:
                    refresh = True
            else:
                refresh = False
        elif not isinstance(refresh, bool):
            raise TypeError("refresh argument should be boolean or `None`")
        self._figure._auto_refresh = refresh

    def refresh(self, force=True):
        """
        Refresh the display.

        Parameters
        ----------
        force : str, optional
            If set to False, refresh() will only have an effect if
            auto refresh is on. If set to True, the display will be
            refreshed whatever the auto refresh setting is set to.
            The default is True.
        """
        if self._figure._auto_refresh or force:
            self._figure.canvas.draw()

    def save(self, filename, dpi=None, transparent=False, adjust_bbox=True,
             max_dpi=300, format=None, **savefig_kwargs):
        """
        Save the current figure to a file.

        Parameters
        ----------

        filename : str or fileobj
            The name of the file to save the plot to. This can be for
            example a PS, EPS, PDF, PNG, JPEG, or SVG file. Note that it
            is also possible to pass file-like object.

        dpi : float, optional
            The output resolution, in dots per inch. If the output file
            is a vector graphics format (such as PS, EPS, PDF or SVG) only
            the image itself will be rasterized. If the output is a PS or
            EPS file and no dpi is specified, the dpi is automatically
            calculated to match the resolution of the image. If this value is
            larger than max_dpi, then dpi is set to max_dpi.

        transparent : str, optional
            Whether to preserve transparency

        adjust_bbox : str, optional
            Auto-adjust the bounding box for the output

        max_dpi : float, optional
            The maximum resolution to output images at. If no maximum is
            wanted, enter None or 0.

        format : str, optional
            By default, APLpy tries to guess the file format based on the
            file extension, but the format can also be specified
            explicitly. Should be one of 'eps', 'ps', 'pdf', 'svg', 'png'.

        **savefig_kwargs
            Additional keywords are passed to Matplotlib's ``savefig()``
        """

        if isinstance(filename, str) and format is None:
            format = os.path.splitext(filename)[1].lower()[1:]

        if dpi is None and format in ['eps', 'ps', 'pdf']:
            width = self.ax.get_position().width * self._figure.get_figwidth()
            interval = self.ax.get_xlim()
            nx = interval[1] - interval[0]
            if max_dpi:
                dpi = np.minimum(nx / width, max_dpi)
            else:
                dpi = nx / width
            log.info("Auto-setting resolution to %g dpi" % dpi)

        if adjust_bbox:
            self._figure.savefig(filename, dpi=dpi, transparent=transparent,
                                 bbox_inches='tight', format=format, **savefig_kwargs)
        else:
            self._figure.savefig(filename, dpi=dpi, transparent=transparent,
                                 format=format, **savefig_kwargs)

    def _initialize_view(self):
        self.ax.set_xlim(-0.5, self._wcs.nx - 0.5)
        self.ax.set_ylim(-0.5, self._wcs.ny - 0.5)

    def _get_invert_default(self):
        return self._figure.apl_grayscale_invert_default

    def _get_colormap_default(self):
        return self._figure.apl_colorscale_cmap_default

    @auto_refresh
    def set_theme(self, theme):
        """
        Set the axes, ticks, grid, and image colors to a certain style
        (experimental).

        Parameters
        ----------

        theme : str
            The theme to use. At the moment, this can be 'pretty' (for
            viewing on-screen) and 'publication' (which makes the ticks
            and grid black, and displays the image in inverted grayscale)
        """

        if theme == 'pretty':
            self.frame.set_color('black')
            self.frame.set_linewidth(1.0)
            if self.ax.coords[self.x].ticks.get_tick_out():
                self.ticks.set_color('black')
            else:
                self.ticks.set_color('white')
            self._figure.apl_grayscale_invert_default = False
            self._figure.apl_colorscale_cmap_default = 'viridis'
            if self.image:
                self.image.set_cmap(cmap=plt.get_cmap('viridis'))
        elif theme == 'publication':
            self.frame.set_color('black')
            self.frame.set_linewidth(1.0)
            self.ticks.set_color('black')
            self.ticks.set_length(7)
            self._figure.apl_grayscale_invert_default = True
            self._figure.apl_colorscale_cmap_default = 'gist_heat'
            if self.image:
                self.image.set_cmap(cmap=plt.get_cmap('gist_yarg'))

    def world2pixel(self, xw, yw, wcs=None):
        """
        Convert world to pixel coordinates.

        Parameters
        ----------
        xw : float or iterable
            x world coordinate
        yw : float or iterable
            y world coordinate

        Returns
        -------
        xp : float or iterable
            x pixel coordinate
        yp : float or iterable
            y pixel coordinate
        """
        if wcs is None:
            wcs = self._wcs
        return wcs.wcs_world2pix(xw, yw, 0)

    def pixel2world(self, xp, yp, wcs=None):
        """
        Convert pixel to world coordinates.

        Parameters
        ----------
        xp : float or iterable
            x pixel coordinate
        yp : float or iterable
            y pixel coordinate

        Returns
        -------
        xw : float or iterable
            x world coordinate
        yw : float or iterable
            y world coordinate
        """
        if wcs is None:
            wcs = self._wcs
        return wcs.wcs_pix2world(xp, yp, 0)

    @auto_refresh
    def add_grid(self):
        """
        Add a coordinate to the current figure.

        Once this method has been run, a grid attribute becomes available,
        and can be used to control the aspect of the grid::

            >>> f = aplpy.FITSFigure(...)
            >>> ...
            >>> f.add_grid()
            >>> f.grid.set_color('white')
            >>> f.grid.set_alpha(0.5)
            >>> ...
        """
        if hasattr(self, 'grid'):
            raise Exception("Grid already exists")
        try:
            self.grid = Grid(self)
            self.grid.show()
        except Exception:
            del self.grid
            raise

    @auto_refresh
    def remove_grid(self):
        """
        Removes the grid from the current figure.
        """
        self.grid.hide()
        del self.grid

    @auto_refresh
    def add_beam(self, *args, **kwargs):
        """
        Add a beam to the current figure.

        Once this method has been run, a beam attribute becomes available,
        and can be used to control the aspect of the beam::

            >>> f = aplpy.FITSFigure(...)
            >>> ...
            >>> f.add_beam()
            >>> f.beam.set_color('white')
            >>> f.beam.set_hatch('+')
            >>> ...

        If more than one beam is added, the beam object becomes a list. In
        this case, to control the aspect of one of the beams, you will need to
        specify the beam index::

            >>> ...
            >>> f.beam[2].set_hatch('/')
            >>> ...
        """

        # Initalize the beam and set parameters
        b = Beam(self)
        b.show(*args, **kwargs)

        if hasattr(self, 'beam'):
            if type(self.beam) is list:
                self.beam.append(b)
            else:
                self.beam = [self.beam, b]
        else:
            self.beam = b

    @auto_refresh
    def remove_beam(self, beam_index=None):
        """
        Removes the beam from the current figure.

        If more than one beam is present, the index of the beam should be
        specified using beam_index=
        """

        if type(self.beam) is list:

            if beam_index is None:
                raise Exception("More than one beam present - use beam_index= to specify which one to remove")
            else:
                b = self.beam.pop(beam_index)
                b._remove()
                del b

            # If only one beam is present, remove containing list
            if len(self.beam) == 1:
                self.beam = self.beam[0]

        else:

            self.beam._remove()
            del self.beam

    @auto_refresh
    def add_scalebar(self, length, *args, **kwargs):
        """
        Add a scalebar to the current figure.

        Once this method has been run, a scalebar attribute becomes
        available, and can be used to control the aspect of the scalebar::

            >>> f = aplpy.FITSFigure(...)
            >>> ...
            >>> f.add_scalebar(0.01) # length has to be specified
            >>> f.scalebar.set_label('100 AU')
            >>> ...

        Parameters
        ----------

        length : float, or quantity
            The length of the scalebar in degrees, an angular quantity, or angular unit

        label : str, optional
            Label to place below the scalebar

        corner : int, optional
            Where to place the scalebar. Acceptable values are:, 'left',
            'right', 'top', 'bottom', 'top left', 'top right', 'bottom
            left' (default), 'bottom right'

        frame : str, optional
            Whether to display a frame behind the scalebar (default is False)

        kwargs
            Additional arguments are passed to the matplotlib Rectangle and
            Text classes. See the matplotlib documentation for more details.
            In cases where the same argument exists for the two objects, the
            argument is passed to both the Text and Rectangle instance.
        """
        if hasattr(self, 'scalebar'):
            raise Exception("Scalebar already exists")
        try:
            self.scalebar = Scalebar(self)
            self.scalebar.show(length, *args, **kwargs)
        except Exception:
            del self.scalebar
            raise

    @auto_refresh
    def remove_scalebar(self):
        """
        Removes the scalebar from the current figure.
        """
        self.scalebar._remove()
        del self.scalebar

    @auto_refresh
    def add_colorbar(self, *args, **kwargs):
        """
        Add a colorbar to the current figure.

        Once this method has been run, a colorbar attribute becomes
        available, and can be used to control the aspect of the colorbar::

            >>> f = aplpy.FITSFigure(...)
            >>> ...
            >>> f.add_colorbar()
            >>> f.colorbar.set_width(0.3)
            >>> f.colorbar.set_location('top')
            >>> ...
        """
        if hasattr(self, 'colorbar'):
            raise Exception("Colorbar already exists")
        if self.image is None:
            raise Exception("No image is shown, so a colorbar cannot be displayed")
        try:
            self.colorbar = Colorbar(self)
            self.colorbar.show(*args, **kwargs)
        except Exception:
            del self.colorbar
            raise

    @auto_refresh
    def remove_colorbar(self):
        """
        Removes the colorbar from the current figure.
        """
        self.colorbar._remove()
        del self.colorbar

    def close(self):
        """
        Close the figure and free up the memory.
        """
        plt.close(self._figure)

    def savefig(self, *args, **kwargs):
        return self.save(*args, **kwargs)



import threading

from functools import wraps

mydata = threading.local()

__all__ = ['auto_refresh', 'fixdocstring']


def auto_refresh(f):

    @wraps(f)
    def wrapper(*args, **kwargs):

        if 'refresh' in kwargs:
            refresh = kwargs.pop('refresh')
        else:
            refresh = True

        # The following is necessary rather than using mydata.nesting = 0 at the
        # start of the file, because doing the latter caused issues with the Django
        # development server.
        mydata.nesting = getattr(mydata, 'nesting', 0) + 1

        try:
            return f(*args, **kwargs)
        finally:
            mydata.nesting -= 1
            if hasattr(args[0], '_figure'):
                if refresh and mydata.nesting == 0 and args[0]._figure._auto_refresh:
                    args[0]._figure.canvas.draw()

    return wrapper


doc = {}

doc['size'] = """size : str or int or float, optional
    The size of the font. This can either be a numeric value (e.g.
    12), giving the size in points, or one of 'xx-small', 'x-small',
    'small', 'medium', 'large', 'x-large', or 'xx-large'.
    """

doc['weight'] = """weight : str or int or float, optional
    The weight (or boldness) of the font. This can either be a numeric
    value in the range 0-1000 or one of 'ultralight', 'light', 'normal',
    'regular', 'book', 'medium', 'roman', 'semibold', 'demibold', 'demi',
    'bold', 'heavy', 'extra bold', 'black'.
    """

doc['stretch'] = """stretch : str or int or float, optional
    The stretching (spacing between letters) for the font. This can either
    be a numeric value in the range 0-1000 or one of 'ultra-condensed',
    'extra-condensed', 'condensed', 'semi-condensed', 'normal',
    'semi-expanded', 'expanded', 'extra-expanded' or 'ultra-expanded'.
    """

doc['family'] = """family : str, optional
    The family of the font to use. This can either be a generic font
    family name, either 'serif', 'sans-serif', 'cursive', 'fantasy', or
    'monospace', or a list of font names in decreasing order of priority.
    """

doc['style'] = """style : str, optional
    The font style. This can be 'normal', 'italic' or 'oblique'.
    """

doc['variant'] = """variant : str, optional
    The font variant. This can be 'normal' or 'small-caps'
    """


def fixdocstring(func):

    lines = func.__doc__.split('\n')

    for i, line in enumerate(lines):
        if 'common:' in line:
            break

    header = lines[:i]

    footer = lines[i + 1:]

    indent = lines[i].index('common:')

    common = []
    for item in lines[i].split(':')[1].split(','):
        if item.strip() in doc:
            common.append(" " * indent + doc[item.strip()].replace('\n', '\n' + " " * indent))

    docstring = "\n".join(header + common + footer)

    func.__doc__ = docstring

    return func



import string
import random
import os

import numpy as np
from astropy.io import fits
from astropy.wcs import WCS


def random_id():
    return ''.join(random.sample(string.ascii_letters + string.digits, 16))


def generate_header(header_file):

    # Read in header
    header = fits.Header.fromtextfile(header_file)

    return header


def generate_data(header_file):

    # Read in header
    header = generate_header(header_file)

    # Find shape of array
    shape = []
    for i in range(header['NAXIS']):
        shape.append(header['NAXIS%i' % (i + 1)])

    # Generate data array
    data = np.zeros(shape[::-1])

    return data


def generate_hdu(header_file):

    # Read in header
    header = generate_header(header_file)

    # Generate data array
    data = generate_data(header_file)

    # Generate primary HDU
    hdu = fits.PrimaryHDU(data=data, header=header)

    return hdu


def generate_wcs(header_file):

    # Read in header
    header = generate_header(header_file)

    # Compute WCS object
    wcs = WCS(header)

    return wcs


def generate_file(header_file, directory):

    # Generate HDU object
    hdu = generate_hdu(header_file)

    # Write out to a temporary file in the specified directory
    filename = os.path.join(directory, random_id() + '.fits')
    hdu.writeto(filename)

    return filename



import numpy as np

from .. import FITSFigure


def test_frame_linewidth():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.frame.set_linewidth(0)
    f.frame.set_linewidth(1)
    f.frame.set_linewidth(10)
    f.close()


def test_frame_color():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.frame.set_color('black')
    f.frame.set_color('#003344')
    f.frame.set_color((1.0, 0.4, 0.3))
    f.close()



from collections import OrderedDict

from matplotlib.contour import ContourSet
from matplotlib.collections import RegularPolyCollection, \
    PatchCollection, CircleCollection, LineCollection

from .regions import ArtistCollection
from .decorators import auto_refresh


class Layers(object):

    def __init__(self):
        pass

    def _layer_type(self, layer):
        if isinstance(self._layers[layer], ContourSet):
            return 'contour'
        elif isinstance(self._layers[layer], RegularPolyCollection):
            return 'collection'
        elif isinstance(self._layers[layer], PatchCollection):
            return 'collection'
        elif isinstance(self._layers[layer], CircleCollection):
            return 'collection'
        elif isinstance(self._layers[layer], LineCollection):
            return 'collection'
        elif isinstance(self._layers[layer], ArtistCollection):
            return 'collection'
        elif hasattr(self._layers[layer], 'remove') and hasattr(self._layers[layer], 'get_visible') and hasattr(self._layers[layer], 'set_visible'):
            return 'collection'
        else:
            raise Exception("Unknown layer type: " +
                            str(type(self._layers[layer])))

    def _initialize_layers(self):

        self._layers = OrderedDict()
        self._contour_counter = 0
        self._vector_counter = 0
        self._scatter_counter = 0
        self._circle_counter = 0
        self._ellipse_counter = 0
        self._rectangle_counter = 0
        self._linelist_counter = 0
        self._region_counter = 0
        self._label_counter = 0
        self._poly_counter = 0

    def list_layers(self):
        """
        Print a list of layers to standard output.
        """

        layers_list = []

        for layer in self._layers:

            layer_type = self._layer_type(layer)

            if layer_type == 'contour':
                visible = self._layers[layer].collections[0].get_visible()
            elif layer_type == 'collection':
                visible = self._layers[layer].get_visible()

            layers_list.append({'name': layer, 'visible': visible})

        n_layers = len(layers_list)
        if n_layers == 0:
            print("\n  There are no layers in this figure")
        else:
            if n_layers == 1:
                print("\n  There is one layer in this figure:\n")
            else:
                print("\n  There are " + str(n_layers) +
                      " layers in this figure:\n")
            for layer in layers_list:
                if layer['visible']:
                    print("   -> " + layer['name'])
                else:
                    print("   -> " + layer['name'] + " (hidden)")

    @auto_refresh
    def remove_layer(self, layer, raise_exception=True):
        """
        Remove a layer.

        Parameters
        ----------
        layer : str
            The name of the layer to remove
        """

        if layer in self._layers:

            layer_type = self._layer_type(layer)

            if layer_type == 'contour':
                for contour in self._layers[layer].collections:
                    contour.remove()
                self._layers.pop(layer)
            elif layer_type == 'collection':
                self._layers[layer].remove()
                self._layers.pop(layer)
                if (layer + '_txt') in self._layers:
                    self._layers[layer + '_txt'].remove()
                    self._layers.pop(layer + '_txt')

        else:

            if raise_exception:
                raise Exception("Layer " + layer + " does not exist")

    @auto_refresh
    def hide_layer(self, layer, raise_exception=True):
        """
        Hide a layer.

        This differs from remove_layer in that if a layer is hidden
        it can be shown again using show_layer.

        Parameters
        ----------
        layer : str
            The name of the layer to hide
        """
        if layer in self._layers:

            layer_type = self._layer_type(layer)

            if layer_type == 'contour':
                for contour in self._layers[layer].collections:
                    contour.set_visible(False)
            elif layer_type == 'collection':
                self._layers[layer].set_visible(False)

        else:

            if raise_exception:
                raise Exception("Layer " + layer + " does not exist")

    @auto_refresh
    def show_layer(self, layer, raise_exception=True):
        """
        Show a layer.

        This shows a layer previously hidden with hide_layer

        Parameters
        ----------
        layer : str
            The name of the layer to show
        """
        if layer in self._layers:

            layer_type = self._layer_type(layer)

            if layer_type == 'contour':
                for contour in self._layers[layer].collections:
                    contour.set_visible(True)
            elif layer_type == 'collection':
                self._layers[layer].set_visible(True)

        else:
            if raise_exception:
                raise Exception("Layer " + layer + " does not exist")

    def get_layer(self, layer, raise_exception=True):
        """
        Return a layer object.

        Parameters
        ----------
        layer : str
            The name of the layer to return
        """
        if layer in self._layers:
            return self._layers[layer]
        else:
            if raise_exception:
                raise Exception("Layer " + layer + " does not exist")



import pytest
import numpy as np

from .. import FITSFigure


def test_layers(capsys):

    f = FITSFigure(np.arange(256).reshape((16, 16)))

    capsys.readouterr()

    # No layers

    f.list_layers()

    captured = capsys.readouterr()
    assert captured.out.strip() == ('There are no layers in this figure')

    # Two layers

    f.show_markers([360., 350., 340.], [-61., -62., -63], color='cyan', layer='markers')
    f.show_markers([360., 350., 340.], [-61., -62., -63], color='cyan', layer='extra_markers')

    capsys.readouterr()

    f.list_layers()

    captured = capsys.readouterr()
    assert captured.out.strip() == ('There are 2 layers in this figure:\n\n'
                                    '   -> markers\n'
                                    '   -> extra_markers')

    # Check removing layers

    f.remove_layer('extra_markers')

    f.list_layers()

    captured = capsys.readouterr()
    assert captured.out.strip() == ('There is one layer in this figure:\n\n'
                                    '   -> markers')

    # Check all layer types and overwriting layers

    f.show_markers([30, 40], [50, 70], layer='markers')

    f.show_circles([360., 350., 340.], [-61., -62., -63], [0.5, 0.4, 0.3], layer='circles')
    f.show_circles([30, 40], [50, 70], [10, 20], layer='circles')

    f.show_ellipses(340., -66., 1.5, 2., 10., layer='ellipses')
    f.show_ellipses(120, 60, 20, 40, 20., layer='ellipses')

    f.show_rectangles([355., 350.], [-71, -72], [0.5, 1], [2, 1], angle=[20, 30], layer='rectangles')
    f.show_rectangles([66, 80], [20, 30], [10, 14], [20, 22], angle=[20, 30], layer='rectangles')

    f.show_arrows([340., 360], [-72, -68], [2, -2], [2, 2], layer='arrows')
    f.show_arrows([340., 360], [-72, -68], [2, -2], [2, 2], layer='arrows')

    poly = np.array([[330, 340, 360], [-65, -61, -63]])
    f.show_polygons([poly], layer='polygons')
    f.show_polygons([poly], layer='polygons')

    f.show_lines([poly], layer='lines')
    f.show_lines([poly], layer='lines')

    f.add_label(350., -66., 'text', layer='label')
    f.add_label(0.4, 0.25, 'text', relative=True, layer='label')

    # FIXME: Uncomment the following once
    # https://github.com/astropy/astropy/pull/8321 is merged into Astropy and in
    # a bug fix release.
    # f.show_contour(layer='contours')

    f.list_layers()

    captured = capsys.readouterr()
    assert captured.out.strip() == ('There are 8 layers in this figure:\n\n'
                                    '   -> markers\n'
                                    '   -> circles\n'
                                    '   -> ellipses\n'
                                    '   -> rectangles\n'
                                    '   -> arrows\n'
                                    '   -> polygons\n'
                                    '   -> lines\n'
                                    '   -> label')

    # Hiding/showing layers

    f.hide_layer('ellipses')
    f.hide_layer('polygons')

    f.list_layers()

    captured = capsys.readouterr()
    assert captured.out.strip() == ('There are 8 layers in this figure:\n\n'
                                    '   -> markers\n'
                                    '   -> circles\n'
                                    '   -> ellipses (hidden)\n'
                                    '   -> rectangles\n'
                                    '   -> arrows\n'
                                    '   -> polygons (hidden)\n'
                                    '   -> lines\n'
                                    '   -> label')

    layer = f.get_layer('circles')
    assert layer.get_visible()

    layer = f.get_layer('ellipses')
    assert not layer.get_visible()

    f.show_layer('ellipses')

    f.list_layers()

    captured = capsys.readouterr()
    assert captured.out.strip() == ('There are 8 layers in this figure:\n\n'
                                    '   -> markers\n'
                                    '   -> circles\n'
                                    '   -> ellipses\n'
                                    '   -> rectangles\n'
                                    '   -> arrows\n'
                                    '   -> polygons (hidden)\n'
                                    '   -> lines\n'
                                    '   -> label')

    layer = f.get_layer('ellipses')
    assert layer.get_visible()


def test_non_existent_layers():

    # Handling non-existent layers

    f = FITSFigure(np.arange(256).reshape((16, 16)))

    with pytest.raises(Exception) as exc:
        f.get_layer('banana')
    assert exc.value.args[0] == 'Layer banana does not exist'

    assert f.get_layer('banana', raise_exception=False) is None

    with pytest.raises(Exception) as exc:
        f.show_layer('banana')
    assert exc.value.args[0] == 'Layer banana does not exist'

    f.show_layer('banana', raise_exception=False)

    with pytest.raises(Exception) as exc:
        f.hide_layer('banana')
    assert exc.value.args[0] == 'Layer banana does not exist'

    f.hide_layer('banana', raise_exception=False)

    with pytest.raises(Exception) as exc:
        f.remove_layer('banana')
    assert exc.value.args[0] == 'Layer banana does not exist'

    f.remove_layer('banana', raise_exception=False)



import os
import sys

from io import BytesIO as StringIO

import pytest
import numpy as np

from .. import FITSFigure

FORMATS = [None, 'png', 'pdf', 'eps', 'ps', 'svg']
ARRAY = np.arange(256).reshape((16, 16))


def is_format(filename, format):
    if isinstance(filename, str):
        f = open(filename, 'rb')
    else:
        f = filename
    if format == 'png':
        return f.read(8) == b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'
    elif format == 'pdf':
        return f.read(4) == b'\x25\x50\x44\x46'
    elif format == 'eps':
        return f.read(23) == b'%!PS-Adobe-3.0 EPSF-3.0'
    elif format == 'ps':
        return f.read(14) == b'%!PS-Adobe-3.0'
    elif format == 'svg':
        from xml.dom import minidom
        return minidom.parse(f).childNodes[-1].attributes['xmlns'].value == 'http://www.w3.org/2000/svg'
    else:
        raise Exception("Unknown format: %s" % format)


@pytest.mark.parametrize(('format'), FORMATS)
def test_write_png(tmpdir, format):
    filename = os.path.join(str(tmpdir), 'test_output.png')
    f = FITSFigure(ARRAY)
    f.show_grayscale()
    try:
        f.save(filename, format=format)
    except TypeError:
        pytest.xfail()
    finally:
        f.close()
    if format is None:
        assert is_format(filename, 'png')
    else:
        assert is_format(filename, format)


@pytest.mark.parametrize(('format'), FORMATS)
def test_write_pdf(tmpdir, format):
    filename = os.path.join(str(tmpdir), 'test_output.pdf')
    f = FITSFigure(ARRAY)
    f.show_grayscale()
    try:
        f.save(filename, format=format)
    except TypeError:
        pytest.xfail()
    finally:
        f.close()
    if format is None:
        assert is_format(filename, 'pdf')
    else:
        assert is_format(filename, format)


@pytest.mark.parametrize(('format'), FORMATS)
def test_write_eps(tmpdir, format):
    filename = os.path.join(str(tmpdir), 'test_output.eps')
    f = FITSFigure(ARRAY)
    f.show_grayscale()
    try:
        f.save(filename, format=format)
    except TypeError:
        pytest.xfail()
    finally:
        f.close()
    if format is None:
        assert is_format(filename, 'eps')
    else:
        assert is_format(filename, format)


@pytest.mark.parametrize(('format'), FORMATS)
def test_write_stringio(tmpdir, format):
    s = StringIO()
    f = FITSFigure(ARRAY)
    f.show_grayscale()
    try:
        f.save(s, format=format)
    except TypeError:
        pytest.xfail()
    finally:
        f.close()
    try:
        s.seek(0)
    except ValueError:
        if format == 'svg' and sys.version_info[:2] >= (3, 3):
            pytest.xfail()
        else:
            raise
    if format is None:
        assert is_format(s, 'png')
    else:
        assert is_format(s, format)



import os
import warnings

import numpy as np
from astropy import log
from astropy.io import fits
from astropy.coordinates import ICRS
from astropy.visualization import AsymmetricPercentileInterval, simple_norm

from reproject import reproject_interp
from reproject.mosaicking import find_optimal_celestial_wcs


def _data_stretch(image, vmin=None, vmax=None, pmin=0.25, pmax=99.75,
                  stretch='linear', vmid=None, exponent=2):

    if vmin is None or vmax is None:
        interval = AsymmetricPercentileInterval(pmin, pmax, n_samples=10000)
        try:
            vmin_auto, vmax_auto = interval.get_limits(image)
        except IndexError:  # no valid values
            vmin_auto = vmax_auto = 0

    if vmin is None:
        log.info("vmin = %10.3e (auto)" % vmin_auto)
        vmin = vmin_auto
    else:
        log.info("vmin = %10.3e" % vmin)

    if vmax is None:
        log.info("vmax = %10.3e (auto)" % vmax_auto)
        vmax = vmax_auto
    else:
        log.info("vmax = %10.3e" % vmax)

    if stretch == 'arcsinh':
        stretch = 'asinh'

    normalizer = simple_norm(image, stretch=stretch, power=exponent,
                             asinh_a=vmid, min_cut=vmin, max_cut=vmax, clip=False)

    data = normalizer(image, clip=True).filled(0)
    data = np.nan_to_num(data)
    data = np.clip(data * 255., 0., 255.)

    return data.astype(np.uint8)


def make_rgb_image(data, output, indices=(0, 1, 2), vmin_r=None, vmax_r=None,
                   pmin_r=0.25, pmax_r=99.75, stretch_r='linear', vmid_r=None,
                   exponent_r=2, vmin_g=None, vmax_g=None, pmin_g=0.25,
                   pmax_g=99.75, stretch_g='linear', vmid_g=None, exponent_g=2,
                   vmin_b=None, vmax_b=None, pmin_b=0.25, pmax_b=99.75,
                   stretch_b='linear', vmid_b=None, exponent_b=2,
                   make_nans_transparent=False, embed_avm_tags=True):
    """
    Make an RGB image from a FITS RGB cube or from three FITS files.

    Parameters
    ----------

    data : str or tuple or list
        If a string, this is the filename of an RGB FITS cube. If a tuple
        or list, this should give the filename of three files to use for
        the red, green, and blue channel.

    output : str
        The output filename. The image type (e.g. PNG, JPEG, TIFF, ...)
        will be determined from the extension. Any image type supported by
        the Python Imaging Library can be used.

    indices : tuple, optional
        If data is the filename of a FITS cube, these indices are the
        positions in the third dimension to use for red, green, and
        blue respectively. The default is to use the first three
        indices.

    vmin_r, vmin_g, vmin_b : float, optional
        Minimum pixel value to use for the red, green, and blue channels.
        If set to None for a given channel, the minimum pixel value for
        that channel is determined using the corresponding pmin_x argument
        (default).

    vmax_r, vmax_g, vmax_b : float, optional
        Maximum pixel value to use for the red, green, and blue channels.
        If set to None for a given channel, the maximum pixel value for
        that channel is determined using the corresponding pmax_x argument
        (default).

    pmin_r, pmin_r, pmin_g : float, optional
        Percentile values used to determine for a given channel the
        minimum pixel value to use for that channel if the corresponding
        vmin_x is set to None. The default is 0.25% for all channels.

    pmax_r, pmax_g, pmax_b : float, optional
        Percentile values used to determine for a given channel the
        maximum pixel value to use for that channel if the corresponding
        vmax_x is set to None. The default is 99.75% for all channels.

    stretch_r, stretch_g, stretch_b : { 'linear', 'log', 'sqrt', 'arcsinh', 'power' }
        The stretch function to use for the different channels.

    vmid_r, vmid_g, vmid_b : float, optional
        Baseline values used for the log and arcsinh stretches. If
        set to None, this is set to zero for log stretches and to
        vmin - (vmax - vmin) / 30. for arcsinh stretches

    exponent_r, exponent_g, exponent_b : float, optional
        If stretch_x is set to 'power', this is the exponent to use.

    make_nans_transparent : bool, optional
        If set AND output is png, will add an alpha layer that sets pixels
        containing a NaN to transparent.

    embed_avm_tags : bool, optional
        Whether to embed AVM tags inside the image - this can only be done for
        JPEG and PNG files, and only if PyAVM is installed.
    """

    try:
        from PIL import Image
    except ImportError:
        try:
            import Image
        except ImportError:
            raise ImportError("The Python Imaging Library (PIL) is required to make an RGB image")

    if isinstance(data, str):

        image = fits.getdata(data)
        image_r = image[indices[0], :, :]
        image_g = image[indices[1], :, :]
        image_b = image[indices[2], :, :]

        # Read in header
        header = fits.getheader(data)

        # Remove information about third dimension
        header['NAXIS'] = 2
        for key in ['NAXIS', 'CTYPE', 'CRPIX', 'CRVAL', 'CUNIT', 'CDELT', 'CROTA']:
            for coord in range(3, 6):
                name = key + str(coord)
                if name in header:
                    header.__delitem__(name)

    elif (type(data) is list or type(data) is tuple) and len(data) == 3:

        filename_r, filename_g, filename_b = data
        image_r = fits.getdata(filename_r)
        image_g = fits.getdata(filename_g)
        image_b = fits.getdata(filename_b)

        # Read in header
        header = fits.getheader(filename_r)

    else:
        raise Exception("data should either be the filename of a FITS cube or a list/tuple of three images")

    # are we making a transparent layer?
    do_alpha = make_nans_transparent and output.lower().endswith('.png')

    if do_alpha:
        log.info("Making alpha layer")

        # initialize alpha layer
        image_alpha = np.empty_like(image_r, dtype=np.uint8)
        image_alpha[:] = 255

        # look for nans in images
        for im in [image_r, image_g, image_b]:
            image_alpha[np.isnan(im)] = 0

    log.info("Red:")
    image_r = Image.fromarray(_data_stretch(image_r, vmin=vmin_r, vmax=vmax_r,
                                            pmin=pmin_r, pmax=pmax_r, stretch=stretch_r,
                                            vmid=vmid_r, exponent=exponent_r))

    log.info("Green:")
    image_g = Image.fromarray(_data_stretch(image_g,
                                            vmin=vmin_g, vmax=vmax_g,
                                            pmin=pmin_g, pmax=pmax_g,
                                            stretch=stretch_g,
                                            vmid=vmid_g,
                                            exponent=exponent_g))

    log.info("Blue:")
    image_b = Image.fromarray(_data_stretch(image_b,
                                            vmin=vmin_b, vmax=vmax_b,
                                            pmin=pmin_b, pmax=pmax_b,
                                            stretch=stretch_b,
                                            vmid=vmid_b,
                                            exponent=exponent_b))

    img = Image.merge("RGB", (image_r, image_g, image_b))

    if do_alpha:
        # convert to RGBA and add alpha layer
        image_alpha = Image.fromarray(image_alpha)
        img.convert("RGBA")
        img.putalpha(image_alpha)

    img = img.transpose(Image.FLIP_TOP_BOTTOM)

    img.save(output)

    if embed_avm_tags:

        from pyavm import AVM

        if output.lower().endswith(('.jpg', '.jpeg', '.png')):
            avm = AVM.from_header(header)
            avm.embed(output, output)
        else:
            warnings.warn("AVM tags will not be embedded in RGB image, as only JPEG and PNG files are supported")


def make_rgb_cube(files, output, north=True):
    """
    Make an RGB data cube from a list of three FITS images.

    This method can read in three FITS files with different
    projections/sizes/resolutions and uses the `reproject
    <https://reproject.readthedocs.io/en/stable/>`_ package to reproject them all
    to the same projection.

    Two files are produced by this function. The first is a three-dimensional
    FITS cube with a filename give by ``output``, where the third dimension
    contains the different channels. The second is a two-dimensional FITS
    image with a filename given by ``output`` with a `_2d` suffix. This file
    contains the mean of the different channels, and is required as input to
    FITSFigure if show_rgb is subsequently used to show a color image
    generated from the FITS cube (to provide the correct WCS information to
    FITSFigure).

    Parameters
    ----------

    files : tuple or list
       A list of the filenames of three FITS filename to reproject.
       The order is red, green, blue.

    output : str
       The filename of the output RGB FITS cube.

    north : bool, optional
        Whether to rotate the image so that north is up. By default, this is
        assumed to be 'north' in the ICRS frame, but you can also pass any
        astropy :class:`~astropy.coordinates.BaseCoordinateFrame` to indicate
        to use the north of that frame.
    """

    # Check that input files exist
    for f in files:
        if not os.path.exists(f):
            raise Exception("File does not exist : " + f)

    if north is not False:
        frame = ICRS() if north is True else north
        auto_rotate = False
    else:
        frame = None
        auto_rotate = True

    # Find optimal WCS and shape based on input images
    wcs, shape = find_optimal_celestial_wcs(files, frame=frame, auto_rotate=auto_rotate)
    header = wcs.to_header()

    # Generate empty datacube
    image_cube = np.zeros((len(files),) + shape, dtype=np.float32)

    # Loop through files and reproject
    for i, filename in enumerate(files):
        image_cube[i, :, :] = reproject_interp(filename, wcs, shape_out=shape)[0]

    # Write out final cube
    fits.writeto(output, image_cube, header, overwrite=True)

    # Write out collapsed version of cube
    fits.writeto(output.replace('.fits', '_2d.fits'),
                 np.mean(image_cube, axis=0), header, overwrite=True)



import numpy as np
from astropy.tests.helper import pytest

from .. import FITSFigure


def test_axis_labels_show_hide():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.axis_labels.hide()
    f.axis_labels.show()
    f.axis_labels.hide_x()
    f.axis_labels.show_x()
    f.axis_labels.hide_y()
    f.axis_labels.show_y()
    f.close()


def test_axis_labels_text():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.axis_labels.set_xtext('x')
    f.axis_labels.set_ytext('y')
    f.close()


def test_axis_labels_pad():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.axis_labels.set_xpad(-1.)
    f.axis_labels.set_ypad(0.5)
    f.close()


def test_axis_labels_position():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.axis_labels.set_xposition('top')
    f.axis_labels.set_xposition('bottom')
    f.axis_labels.set_yposition('right')
    f.axis_labels.set_yposition('left')
    f.close()


def test_axis_labels_position_invalid():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    with pytest.raises(ValueError):
        f.axis_labels.set_xposition('right')
    with pytest.raises(ValueError):
        f.axis_labels.set_xposition('left')
    with pytest.raises(ValueError):
        f.axis_labels.set_yposition('top')
    with pytest.raises(ValueError):
        f.axis_labels.set_yposition('bottom')
    f.close()


def test_axis_labels_font():
    data = np.zeros((16, 16))
    f = FITSFigure(data)
    f.axis_labels.set_font(size='small', weight='bold', stretch='normal',
                           family='serif', style='normal', variant='normal')
    f.close()



import warnings

import matplotlib.axes as maxes
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.font_manager import FontProperties
from matplotlib.ticker import LogFormatterMathtext

from .decorators import auto_refresh, fixdocstring


class Colorbar(object):

    def __init__(self, parent):
        self._figure = parent._figure
        self._colorbar_axes = None
        self._parent = parent

        self._base_settings = {}
        self._ticklabel_fontproperties = FontProperties()
        self._axislabel_fontproperties = FontProperties()

    @auto_refresh
    def show(self, location='right', width=0.2, pad=0.05, ticks=None,
             labels=True, log_format=False, box=None,
             box_orientation='vertical', axis_label_text=None,
             axis_label_rotation=None, axis_label_pad=5):
        """
        Show a colorbar on the side of the image.

        Parameters
        ----------

        location : str, optional
            Where to place the colorbar. Should be one of 'left', 'right',
            'top', 'bottom'.

        width : float, optional
            The width of the colorbar relative to the canvas size.

        pad : float, optional
            The spacing between the colorbar and the image relative to the
            canvas size.

        ticks : list, optional
            The position of the ticks on the colorbar.

        labels : bool, optional
            Whether to show numerical labels.

        log_format : bool, optional
            Whether to format ticks in exponential notation

        box : list, optional
            A custom box within which to place the colorbar. This should
            be in the form [xmin, ymin, dx, dy] and be in relative figure
            units. This overrides the location argument.

        box_orientation str, optional
            The orientation of the colorbar within the box. Can be
            'horizontal' or 'vertical'

        axis_label_text str, optional
            Optional text label of the colorbar.
        """

        self._base_settings['location'] = location
        self._base_settings['width'] = width
        self._base_settings['pad'] = pad
        self._base_settings['ticks'] = ticks
        self._base_settings['labels'] = labels
        self._base_settings['log_format'] = log_format
        self._base_settings['box'] = box
        self._base_settings['box_orientation'] = box_orientation
        self._base_settings['axis_label_text'] = axis_label_text
        self._base_settings['axis_label_rotation'] = axis_label_rotation
        self._base_settings['axis_label_pad'] = axis_label_pad

        if self._parent.image:

            if self._colorbar_axes:
                self._figure.delaxes(self._colorbar_axes)

            if box is None:

                divider = make_axes_locatable(self._parent.ax)

                if location == 'right':
                    self._colorbar_axes = divider.new_horizontal(size=width, pad=pad, axes_class=maxes.Axes)
                    orientation = 'vertical'
                elif location == 'top':
                    self._colorbar_axes = divider.new_vertical(size=width, pad=pad, axes_class=maxes.Axes)
                    orientation = 'horizontal'
                elif location == 'left':
                    warnings.warn("Left colorbar not fully implemented")
                    self._colorbar_axes = divider.new_horizontal(size=width, pad=pad, pack_start=True, axes_class=maxes.Axes)
                    locator = divider.new_locator(nx=0, ny=0)
                    self._colorbar_axes.set_axes_locator(locator)
                    orientation = 'vertical'
                elif location == 'bottom':
                    warnings.warn("Bottom colorbar not fully implemented")
                    self._colorbar_axes = divider.new_vertical(size=width, pad=pad, pack_start=True, axes_class=maxes.Axes)
                    locator = divider.new_locator(nx=0, ny=0)
                    self._colorbar_axes.set_axes_locator(locator)
                    orientation = 'horizontal'
                else:
                    raise Exception("location should be one of: right/top")

                self._figure.add_axes(self._colorbar_axes)

            else:

                self._colorbar_axes = self._figure.add_axes(box)
                orientation = box_orientation

            if log_format:
                format = LogFormatterMathtext()
            else:
                format = None

            self._colorbar = self._figure.colorbar(self._parent.image, cax=self._colorbar_axes,
                                                   orientation=orientation, format=format,
                                                   ticks=ticks)
            if axis_label_text:
                if axis_label_rotation:
                    self._colorbar.set_label(axis_label_text, rotation=axis_label_rotation)
                else:
                    self._colorbar.set_label(axis_label_text)

            if location == 'right':
                for tick in self._colorbar_axes.yaxis.get_major_ticks():
                    tick.tick1line.set_visible(True)
                    tick.tick2line.set_visible(True)
                    tick.label1.set_visible(False)
                    tick.label2.set_visible(labels)
                self._colorbar_axes.yaxis.set_label_position('right')
                self._colorbar_axes.yaxis.labelpad = axis_label_pad
            elif location == 'top':
                for tick in self._colorbar_axes.xaxis.get_major_ticks():
                    tick.tick1line.set_visible(True)
                    tick.tick2line.set_visible(True)
                    tick.label1.set_visible(False)
                    tick.label2.set_visible(labels)
                self._colorbar_axes.xaxis.set_label_position('top')
                self._colorbar_axes.xaxis.labelpad = axis_label_pad
            elif location == 'left':
                for tick in self._colorbar_axes.yaxis.get_major_ticks():
                    tick.tick1line.set_visible(True)
                    tick.tick2line.set_visible(True)
                    tick.label1.set_visible(labels)
                    tick.label2.set_visible(False)
                self._colorbar_axes.yaxis.set_label_position('left')
                self._colorbar_axes.yaxis.labelpad = axis_label_pad
            elif location == 'bottom':
                for tick in self._colorbar_axes.xaxis.get_major_ticks():
                    tick.tick1line.set_visible(True)
                    tick.tick2line.set_visible(True)
                    tick.label1.set_visible(labels)
                    tick.label2.set_visible(False)
                self._colorbar_axes.xaxis.set_label_position('bottom')
                self._colorbar_axes.xaxis.labelpad = axis_label_pad

        else:

            warnings.warn("No image is shown, therefore, no colorbar will be plotted")

    @auto_refresh
    def update(self):
        if self._colorbar_axes:
            self.show(**self._base_settings)

    @auto_refresh
    def hide(self):
        self._figure.delaxes(self._colorbar_axes)
        self._colorbar_axes = None

    @auto_refresh
    def _remove(self):
        self._figure.delaxes(self._colorbar_axes)

    # LOCATION AND SIZE

    @auto_refresh
    def set_location(self, location):
        """
        Set the location of the colorbar.

        Should be one of 'left', 'right', 'top', 'bottom'.
        """
        self._base_settings['location'] = location
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_width(self, width):
        """
        Set the width of the colorbar relative to the canvas size.
        """
        self._base_settings['width'] = width
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_pad(self, pad):
        """
        Set the spacing between the colorbar and the image relative to the
        canvas size.
        """
        self._base_settings['pad'] = pad
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_ticks(self, ticks):
        """
        Set the position of the ticks on the colorbar.
        """
        self._base_settings['ticks'] = ticks
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_labels(self, labels):
        """
        Set whether to show numerical labels.
        """
        self._base_settings['labels'] = labels
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_box(self, box, box_orientation='vertical'):
        """
        Set the box within which to place the colorbar.

        This should be in the form [xmin, ymin, dx, dy] and be in relative
        figure units. The orientation of the colorbar within the box can be
        controlled with the box_orientation argument.
        """
        self._base_settings['box'] = box
        self._base_settings['box_orientation'] = box_orientation
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_axis_label_text(self, axis_label_text):
        """
        Set the colorbar label text.
        """
        self._base_settings['axis_label_text'] = axis_label_text
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_axis_label_rotation(self, axis_label_rotation):
        """
        Set the colorbar label rotation.
        """
        self._base_settings['axis_label_rotation'] = axis_label_rotation
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    @auto_refresh
    def set_axis_label_pad(self, axis_label_pad):
        """
        Set the colorbar label displacement, in points.
        """
        self._base_settings['axis_label_pad'] = axis_label_pad
        self.show(**self._base_settings)
        self.set_font(fontproperties=self._ticklabel_fontproperties)
        self.set_axis_label_font(fontproperties=self._axislabel_fontproperties)

    # FONT PROPERTIES

    @auto_refresh
    def set_label_properties(self, *args, **kwargs):
        warnings.warn("set_label_properties is deprecated - use set_font instead", DeprecationWarning)
        self.set_font(*args, **kwargs)

    @auto_refresh
    @fixdocstring
    def set_font(self, family=None, style=None, variant=None, stretch=None,
                 weight=None, size=None, fontproperties=None):
        """
        Set the font of the tick labels.

        Parameters
        ----------

        common: family, style, variant, stretch, weight, size, fontproperties

        Notes
        -----

        Default values are set by matplotlib or previously set values if
        set_font has already been called. Global default values can be set by
        editing the matplotlibrc file.
        """

        if family:
            self._ticklabel_fontproperties.set_family(family)

        if style:
            self._ticklabel_fontproperties.set_style(style)

        if variant:
            self._ticklabel_fontproperties.set_variant(variant)

        if stretch:
            self._ticklabel_fontproperties.set_stretch(stretch)

        if weight:
            self._ticklabel_fontproperties.set_weight(weight)

        if size:
            self._ticklabel_fontproperties.set_size(size)

        if fontproperties:
            self._ticklabel_fontproperties = fontproperties

        # Update the tick label font properties
        for label in self._colorbar_axes.get_xticklabels():
            label.set_fontproperties(self._ticklabel_fontproperties)
        for label in self._colorbar_axes.get_yticklabels():
            label.set_fontproperties(self._ticklabel_fontproperties)

        # Also update the offset text font properties
        label = self._colorbar_axes.xaxis.get_offset_text()
        label.set_fontproperties(self._ticklabel_fontproperties)
        label = self._colorbar_axes.yaxis.get_offset_text()
        label.set_fontproperties(self._ticklabel_fontproperties)

    @auto_refresh
    @fixdocstring
    def set_axis_label_font(self, family=None, style=None, variant=None,
                            stretch=None, weight=None, size=None,
                            fontproperties=None):
        """
        Set the font of the tick labels.

        Parameters
        ----------

        common: family, style, variant, stretch, weight, size, fontproperties

        Notes
        -----

        Default values are set by matplotlib or previously set values if
        set_font has already been called. Global default values can be set by
        editing the matplotlibrc file.
        """

        if family:
            self._axislabel_fontproperties.set_family(family)

        if style:
            self._axislabel_fontproperties.set_style(style)

        if variant:
            self._axislabel_fontproperties.set_variant(variant)

        if stretch:
            self._axislabel_fontproperties.set_stretch(stretch)

        if weight:
            self._axislabel_fontproperties.set_weight(weight)

        if size:
            self._axislabel_fontproperties.set_size(size)

        if fontproperties:
            self._axislabel_fontproperties = fontproperties

        # Update the label font properties
        label = self._colorbar_axes.xaxis.get_label()
        label.set_fontproperties(self._axislabel_fontproperties)
        label = self._colorbar_axes.yaxis.get_label()
        label.set_fontproperties(self._axislabel_fontproperties)

    # FRAME PROPERTIES

    @auto_refresh
    def set_frame_linewidth(self, linewidth):
        """
        Set the linewidth of the colorbar frame, in points.
        """
        warnings.warn("This method is not functional at this time")
        for key in self._colorbar_axes.spines:
            self._colorbar_axes.spines[key].set_linewidth(linewidth)

    @auto_refresh
    def set_frame_color(self, color):
        """
        Set the color of the colorbar frame, in points.
        """
        warnings.warn("This method is not functional at this time")
        for key in self._colorbar_axes.spines:
            self._colorbar_axes.spines[key].set_edgecolor(color)



import warnings

from matplotlib.offsetbox import AnchoredOffsetbox, AuxTransformBox
from matplotlib.patches import Ellipse
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

import numpy as np
from matplotlib.font_manager import FontProperties

from astropy import units as u
from astropy.wcs.utils import proj_plane_pixel_scales

from .decorators import auto_refresh

corners = {}
corners['top right'] = 1
corners['top left'] = 2
corners['bottom left'] = 3
corners['bottom right'] = 4
corners['right'] = 5
corners['left'] = 6
corners['bottom'] = 8
corners['top'] = 9


class Scalebar(object):

    def __init__(self, parent):

        # Retrieve info from parent figure
        self._ax = parent.ax
        self._wcs = parent._wcs
        self._figure = parent._figure
        self._dimensions = [parent.x, parent.y]

        # Initialize settings
        self._base_settings = {}
        self._scalebar_settings = {}
        self._label_settings = {}
        self._label_settings['fontproperties'] = FontProperties()

    # LAYOUT

    @auto_refresh
    def show(self, length, label=None, corner='bottom right', frame=False,
             borderpad=0.4, pad=0.5, **kwargs):
        """
        Overlay a scale bar on the image.

        Parameters
        ----------

        length : float, or quantity
            The length of the scalebar in degrees, an angular quantity, or angular unit

        label : str, optional
            Label to place below the scalebar

        corner : int, optional
            Where to place the scalebar. Acceptable values are:, 'left',
            'right', 'top', 'bottom', 'top left', 'top right', 'bottom
            left' (default), 'bottom right'

        frame : str, optional
            Whether to display a frame behind the scalebar (default is False)

        kwargs
            Additional arguments are passed to the matplotlib Rectangle and
            Text classes. See the matplotlib documentation for more details.
            In cases where the same argument exists for the two objects, the
            argument is passed to both the Text and Rectangle instance.
        """

        self._length = length
        self._base_settings['corner'] = corner
        self._base_settings['frame'] = frame
        self._base_settings['borderpad'] = borderpad
        self._base_settings['pad'] = pad

        if isinstance(length, u.Quantity):
            length = length.to(u.degree).value
        elif isinstance(length, u.Unit):
            length = length.to(u.degree)

        if self._wcs.is_celestial:
            pix_scale = proj_plane_pixel_scales(self._wcs)
            sx = pix_scale[self._dimensions[0]]
            sy = pix_scale[self._dimensions[1]]
            degrees_per_pixel = np.sqrt(sx * sy)
        else:
            raise ValueError("Cannot show scalebar when WCS is not celestial")

        length = length / degrees_per_pixel

        try:
            self._scalebar.remove()
        except Exception:
            pass

        if isinstance(corner, str):
            corner = corners[corner]

        self._scalebar = AnchoredSizeBar(self._ax.transData, length, label,
                                         corner, pad=pad, borderpad=borderpad,
                                         sep=5, frameon=frame)

        self._ax.add_artist(self._scalebar)

        self.set(**kwargs)

    @auto_refresh
    def _remove(self):
        self._scalebar.remove()

    @auto_refresh
    def hide(self):
        """
        Hide the scalebar.
        """
        try:
            self._scalebar.remove()
        except Exception:
            pass

    @auto_refresh
    def set_length(self, length):
        """
        Set the length of the scale bar.
        """
        self.show(length, **self._base_settings)
        self._set_scalebar_properties(**self._scalebar_settings)
        self._set_label_properties(**self._scalebar_settings)

    @auto_refresh
    def set_label(self, label):
        """
        Set the label of the scale bar.
        """
        self._set_label_properties(text=label)

    @auto_refresh
    def set_corner(self, corner):
        """
        Set where to place the scalebar.

        Acceptable values are 'left', 'right', 'top', 'bottom', 'top left',
        'top right', 'bottom left' (default), and 'bottom right'.
        """
        self._base_settings['corner'] = corner
        self.show(self._length, **self._base_settings)
        self._set_scalebar_properties(**self._scalebar_settings)
        self._set_label_properties(**self._scalebar_settings)

    @auto_refresh
    def set_frame(self, frame):
        """
        Set whether to display a frame around the scalebar.
        """
        self._base_settings['frame'] = frame
        self.show(self._length, **self._base_settings)
        self._set_scalebar_properties(**self._scalebar_settings)
        self._set_label_properties(**self._scalebar_settings)

    # APPEARANCE

    @auto_refresh
    def set_linewidth(self, linewidth):
        """
        Set the linewidth of the scalebar, in points.
        """
        self._set_scalebar_properties(linewidth=linewidth)

    @auto_refresh
    def set_linestyle(self, linestyle):
        """
        Set the linestyle of the scalebar.

        Should be one of 'solid', 'dashed', 'dashdot', or 'dotted'.
        """
        self._set_scalebar_properties(linestyle=linestyle)

    @auto_refresh
    def set_alpha(self, alpha):
        """
        Set the alpha value (transparency).

        This should be a floating point value between 0 and 1.
        """
        self._set_scalebar_properties(alpha=alpha)
        self._set_label_properties(alpha=alpha)

    @auto_refresh
    def set_color(self, color):
        """
        Set the label and scalebar color.
        """
        self._set_scalebar_properties(color=color)
        self._set_label_properties(color=color)

    @auto_refresh
    def set_font(self, family=None, style=None, variant=None, stretch=None,
                 weight=None, size=None, fontproperties=None):
        """
        Set the font of the tick labels

        Parameters
        ----------

        common: family, style, variant, stretch, weight, size, fontproperties

        Notes
        -----

        Default values are set by matplotlib or previously set values if
        set_font has already been called. Global default values can be set by
        editing the matplotlibrc file.
        """

        if family:
            self._label_settings['fontproperties'].set_family(family)

        if style:
            self._label_settings['fontproperties'].set_style(style)

        if variant:
            self._label_settings['fontproperties'].set_variant(variant)

        if stretch:
            self._label_settings['fontproperties'].set_stretch(stretch)

        if weight:
            self._label_settings['fontproperties'].set_weight(weight)

        if size:
            self._label_settings['fontproperties'].set_size(size)

        if fontproperties:
            self._label_settings['fontproperties'] = fontproperties

        self._set_label_properties(fontproperties=self._label_settings['fontproperties'])

    @auto_refresh
    def _set_label_properties(self, **kwargs):
        """
        Modify the scalebar label properties.

        All arguments are passed to the matplotlib Text class. See the
        matplotlib documentation for more details.
        """
        for kwarg, val in kwargs.items():
            try:
                # Only set attributes that exist
                kvpair = {kwarg: val}
                self._scalebar.txt_label.get_children()[0].set(**kvpair)
                self._label_settings[kwarg] = val
            except AttributeError:
                warnings.warn("Text labels do not have attribute {0}. Skipping.".format(kwarg))

    @auto_refresh
    def _set_scalebar_properties(self, **kwargs):
        """
        Modify the scalebar properties.

        All arguments are passed to the matplotlib Rectangle class. See the
        matplotlib documentation for more details.
        """
        for kwarg, val in kwargs.items():
            try:
                kvpair = {kwarg: val}
                self._scalebar.size_bar.get_children()[0].set(**kvpair)
                self._scalebar_settings[kwarg] = val
            except AttributeError:
                warnings.warn("Scalebar does not have attribute {0}. Skipping.".format(kwarg))

    @auto_refresh
    def set(self, **kwargs):
        """
        Modify the scalebar and scalebar properties.

        All arguments are passed to the matplotlib Rectangle and Text classes.
        See the matplotlib documentation for more details. In cases where the
        same argument exists for the two objects, the argument is passed to
        both the Text and Rectangle instance.
        """
        for kwarg in kwargs:
            kwargs_single = {kwarg: kwargs[kwarg]}
            try:
                self._set_label_properties(**kwargs_single)
            except (AttributeError, TypeError):
                pass
            try:
                self._set_scalebar_properties(**kwargs_single)
            except (AttributeError, TypeError):
                pass

    # DEPRECATED

    @auto_refresh
    def set_font_family(self, family):
        warnings.warn("scalebar.set_font_family is deprecated - use scalebar.set_font instead", DeprecationWarning)
        self.set_font(family=family)

    @auto_refresh
    def set_font_weight(self, weight):
        warnings.warn("scalebar.set_font_weight is deprecated - use scalebar.set_font instead", DeprecationWarning)
        self.set_font(weight=weight)

    @auto_refresh
    def set_font_size(self, size):
        warnings.warn("scalebar.set_font_size is deprecated - use scalebar.set_font instead", DeprecationWarning)
        self.set_font(size=size)

    @auto_refresh
    def set_font_style(self, style):
        warnings.warn("scalebar.set_font_style is deprecated - use scalebar.set_font instead", DeprecationWarning)
        self.set_font(style=style)


# For backward-compatibility
ScaleBar = Scalebar


class Beam(object):

    def __init__(self, parent):

        # Retrieve info from parent figure
        self._figure = parent._figure
        self._header = parent._header
        self._ax = parent.ax
        self._wcs = parent._wcs
        self._dimensions = [parent.x, parent.y]

        # Initialize settings
        self._base_settings = {}
        self._beam_settings = {}

    # LAYOUT

    @auto_refresh
    def show(self, major='BMAJ', minor='BMIN', angle='BPA',
             corner='bottom left', frame=False, borderpad=0.4, pad=0.5,
             **kwargs):
        """
        Display the beam shape and size for the primary image.

        By default, this method will search for the BMAJ, BMIN, and BPA
        keywords in the FITS header to set the major and minor axes and the
        position angle on the sky.

        Parameters
        ----------

        major : float, quantity or unit, optional
            Major axis of the beam in degrees or an angular quantity (overrides
            BMAJ if present)

        minor : float, quantity or unit, optional
            Minor axis of the beam in degrees or an angular quantity (overrides
            BMIN if present)

        angle : float, quantity or unit, optional
            Position angle of the beam on the sky in degrees or an angular
            quantity (overrides BPA if present) in the anticlockwise direction.

        corner : str, optional
            The beam location. Acceptable values are 'left', 'right',
            'top', 'bottom', 'top left', 'top right', 'bottom left'
            (default), and 'bottom right'.

        frame : str, optional
            Whether to display a frame behind the beam (default is False)

        kwargs
            Additional arguments are passed to the matplotlib Ellipse class.
            See the matplotlib documentation for more details.
        """

        if isinstance(major, str):
            major = self._header[major]

        if isinstance(minor, str):
            minor = self._header[minor]

        if isinstance(angle, str):
            angle = self._header[angle]

        if isinstance(major, u.Quantity):
            major = major.to(u.degree).value
        elif isinstance(major, u.Unit):
            major = major.to(u.degree)

        if isinstance(minor, u.Quantity):
            minor = minor.to(u.degree).value
        elif isinstance(minor, u.Unit):
            minor = minor.to(u.degree)

        if isinstance(angle, u.Quantity):
            angle = angle.to(u.degree).value
        elif isinstance(angle, u.Unit):
            angle = angle.to(u.degree)

        if self._wcs.is_celestial:
            pix_scale = proj_plane_pixel_scales(self._wcs)
            sx = pix_scale[self._dimensions[0]]
            sy = pix_scale[self._dimensions[1]]
            degrees_per_pixel = np.sqrt(sx * sy)
        else:
            raise ValueError("Cannot show beam when WCS is not celestial")

        self._base_settings['minor'] = minor
        self._base_settings['major'] = major
        self._base_settings['angle'] = angle
        self._base_settings['corner'] = corner
        self._base_settings['frame'] = frame
        self._base_settings['borderpad'] = borderpad
        self._base_settings['pad'] = pad

        minor /= degrees_per_pixel
        major /= degrees_per_pixel

        try:
            self._beam.remove()
        except Exception:
            pass

        if isinstance(corner, str):
            corner = corners[corner]

        aux_tr_box = AuxTransformBox(self._ax.transData)
        self._ellipse = Ellipse((0, 0), width=minor, height=major, angle=angle, **kwargs)
        aux_tr_box.add_artist(self._ellipse)

        self._beam = AnchoredOffsetbox(
            child=aux_tr_box,
            loc=corner,
            pad=pad,
            borderpad=borderpad,
            frameon=frame)

        self._ax.add_artist(self._beam)

        self.set(**kwargs)

    @auto_refresh
    def _remove(self):
        self._beam.remove()

    @auto_refresh
    def hide(self):
        """
        Hide the beam
        """
        try:
            self._beam.remove()
        except Exception:
            pass

    @auto_refresh
    def set_major(self, major):
        """
        Set the major axis of the beam, in degrees.
        """
        self._base_settings['major'] = major
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    @auto_refresh
    def set_minor(self, minor):
        """
        Set the minor axis of the beam, in degrees.
        """
        self._base_settings['minor'] = minor
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    @auto_refresh
    def set_angle(self, angle):
        """
        Set the position angle of the beam on the sky, in degrees.
        """
        self._base_settings['angle'] = angle
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    @auto_refresh
    def set_corner(self, corner):
        """
        Set the beam location.

        Acceptable values are 'left', 'right', 'top', 'bottom', 'top left',
        'top right', 'bottom left' (default), and 'bottom right'.
        """
        self._base_settings['corner'] = corner
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    @auto_refresh
    def set_frame(self, frame):
        """
        Set whether to display a frame around the beam.
        """
        self._base_settings['frame'] = frame
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    @auto_refresh
    def set_borderpad(self, borderpad):
        """
        Set the amount of padding within the beam object, relative to the
        canvas size.
        """
        self._base_settings['borderpad'] = borderpad
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    @auto_refresh
    def set_pad(self, pad):
        """
        Set the amount of padding between the beam object and the image
        corner/edge, relative to the canvas size.
        """
        self._base_settings['pad'] = pad
        self.show(**self._base_settings)
        self.set(**self._beam_settings)

    # APPEARANCE

    @auto_refresh
    def set_alpha(self, alpha):
        """
        Set the alpha value (transparency).

        This should be a floating point value between 0 and 1.
        """
        self.set(alpha=alpha)

    @auto_refresh
    def set_color(self, color):
        """
        Set the beam color.
        """
        self.set(color=color)

    @auto_refresh
    def set_edgecolor(self, edgecolor):
        """
        Set the color for the edge of the beam.
        """
        self.set(edgecolor=edgecolor)

    @auto_refresh
    def set_facecolor(self, facecolor):
        """
        Set the color for the interior of the beam.
        """
        self.set(facecolor=facecolor)

    @auto_refresh
    def set_linestyle(self, linestyle):
        """
        Set the line style for the edge of the beam.

        This should be one of 'solid', 'dashed', 'dashdot', or 'dotted'.
        """
        self.set(linestyle=linestyle)

    @auto_refresh
    def set_linewidth(self, linewidth):
        """
        Set the line width for the edge of the beam, in points.
        """
        self.set(linewidth=linewidth)

    @auto_refresh
    def set_hatch(self, hatch):
        """
        Set the hatch pattern.

        This should be one of '/', '\', '|', '-', '+', 'x', 'o', 'O', '.', or
        '*'.
        """
        self.set(hatch=hatch)

    @auto_refresh
    def set(self, **kwargs):
        """
        Modify the beam properties. All arguments are passed to the matplotlib
        Ellipse class. See the matplotlib documentation for more details.
        """
        for kwarg in kwargs:
            self._beam_settings[kwarg] = kwargs[kwarg]
        self._ellipse.set(**kwargs)




def slice_hypercube(data, header, dimensions=[0, 1], slices=[]):
    """
    Extract a slice from an n-dimensional HDU data/header pair, and return the
    new data (without changing the header).
    """

    if type(slices) is int:
        slices = (slices, )
    else:
        slices = slices[:]

    shape = data.shape

    if len(shape) < 2:

        raise Exception("FITS file does not have enough dimensions")

    elif len(shape) == 2:
        wcsaxes_slices = ('x', 'y')

        if dimensions[1] < dimensions[0]:
            data = data.transpose()
            wcsaxes_slices = ('y', 'x')

        return data, wcsaxes_slices

    else:

        if slices:
            wcsaxes_slices = slices[:]

            if dimensions[0] < dimensions[1]:
                slices.insert(dimensions[0], slice(None, None, None))
                slices.insert(dimensions[1], slice(None, None, None))
                wcsaxes_slices.insert(dimensions[0], 'x')
                wcsaxes_slices.insert(dimensions[1], 'y')
            else:
                slices.insert(dimensions[1], slice(None, None, None))
                slices.insert(dimensions[0], slice(None, None, None))
                wcsaxes_slices.insert(dimensions[1], 'y')
                wcsaxes_slices.insert(dimensions[0], 'x')

            if type(slices) is list:
                slices = tuple(slices)
                wcsaxes_slices = tuple(wcsaxes_slices)

            data = data[slices[::-1]]

            if dimensions[1] < dimensions[0]:
                data = data.transpose()

        else:

            message = """
    Attempted to read in %i-dimensional FITS cube, but
    dimensions and slices were not specified. Please specify these
    using the dimensions= and slices= argument. The cube dimensions
    are:\n\n""" % len(shape)

            for i in range(1, len(shape) + 1):

                message += " " * 10
                message += " %i %s %i\n" % (i - 1,
                                            header["CTYPE%i" % i],
                                            header["NAXIS%i" % i])

            raise Exception(message)

        return data, wcsaxes_slices



import numpy as np

from .decorators import auto_refresh


class Grid(object):

    @auto_refresh
    def __init__(self, parent):

        # Save axes and wcs information
        self.ax = parent.ax
        self._wcs = parent._wcs
        self._figure = parent._figure
        self.x = parent.x
        self.y = parent.y
        self.x_unit = self._wcs.wcs.cunit[self.x]
        self.y_unit = self._wcs.wcs.cunit[self.y]
        self.grid_type = parent.grid_type

        # Set defaults
        self.default_color = 'white'
        self.default_alpha = 0.5

    @auto_refresh
    def set_xspacing(self, xspacing):
        """
        Set the grid line spacing in the longitudinal direction

        Parameters
        ----------
        xspacing : float or str
            The spacing in the longitudinal direction. To set the spacing
            to be the same as the ticks, set this to 'tick'
        """
        if xspacing == 'tick':
            self.ax.coords[self.x].grid(grid_type=self.grid_type)
        elif np.isreal(xspacing):
            self.ax.coords[self.x].set_ticks(spacing=xspacing * self.x_unit)
        else:
            raise ValueError("Grid spacing should be a scalar or 'tick'")

    @auto_refresh
    def set_yspacing(self, yspacing):
        """
        Set the grid line spacing in the latitudinal direction

        Parameters
        ----------
        yspacing : { float, str }
            The spacing in the latitudinal direction. To set the spacing
            to be the same as the ticks, set this to 'tick'
        """
        if yspacing == 'tick':
            self.ax.coords[self.y].grid(grid_type=self.grid_type)
        elif np.isreal(yspacing):
            self.ax.coords[self.y].set_ticks(spacing=yspacing * self.y_unit)
        else:
            raise ValueError("Grid spacing should be a scalar or 'tick'")

    @auto_refresh
    def set_color(self, color):
        """
        Set the color of the grid lines

        Parameters
        ----------
        color : str
            The color of the grid lines
        """
        self.default_color = color
        self.ax.coords[self.x].grid(color=color, grid_type=self.grid_type)
        self.ax.coords[self.y].grid(color=color, grid_type=self.grid_type)

    @auto_refresh
    def set_alpha(self, alpha):
        """
        Set the alpha (transparency) of the grid lines

        Parameters
        ----------
        alpha : float
            The alpha value of the grid. This should be a floating
            point value between 0 and 1, where 0 is completely
            transparent, and 1 is completely opaque.
        """
        self.default_alpha = alpha
        self.ax.coords[self.x].grid(alpha=alpha, grid_type=self.grid_type)
        self.ax.coords[self.y].grid(alpha=alpha, grid_type=self.grid_type)

    @auto_refresh
    def set_linewidth(self, linewidth):
        self.ax.coords[self.x].grid(linewidth=linewidth, grid_type=self.grid_type)
        self.ax.coords[self.y].grid(linewidth=linewidth, grid_type=self.grid_type)

    @auto_refresh
    def set_linestyle(self, linestyle):
        self.ax.coords[self.x].grid(linestyle=linestyle, grid_type=self.grid_type)
        self.ax.coords[self.y].grid(linestyle=linestyle, grid_type=self.grid_type)

    @auto_refresh
    def show(self):
        self.ax.coords[self.x].grid(grid_type=self.grid_type,
                                    color=self.default_color,
                                    alpha=self.default_alpha)
        self.ax.coords[self.y].grid(grid_type=self.grid_type,
                                    color=self.default_color,
                                    alpha=self.default_alpha)

    @auto_refresh
    def hide(self):
        self.ax.coords[self.x].grid(draw_grid=False)
        self.ax.coords[self.y].grid(draw_grid=False)



import numpy as np

from ..core import FITSFigure

from .test_images import BaseImageTests
from .figures import figure_test

x = np.linspace(-1., 1., 10)
y = np.linspace(-1., 1., 10)
X, Y = np.meshgrid(x, y)

np.random.seed(12345)
IMAGE = np.random.random((10, 10))
PDATA = np.arange(100).reshape((10, 10)) / 50. + 0.5
ADATA = np.degrees(np.arctan2(Y, X))


class TestVectors(BaseImageTests):

    @figure_test
    def test_default(self):
        f = FITSFigure(IMAGE, figsize=(4, 4))
        f.show_grayscale()
        f.show_vectors(PDATA, ADATA, color='orange')
        return f

    @figure_test
    def test_step_scale(self):
        f = FITSFigure(IMAGE, figsize=(4, 4))
        f.show_grayscale()
        f.show_vectors(PDATA, ADATA, step=2, scale=0.8, color='orange')
        return f



# The simple_norm function in astropy is missing a control for the log
# scaling as described in https://github.com/astropy/astropy/issues/8432,
# so for now we include a copy of the fixed function here.


from astropy.visualization.interval import (PercentileInterval,
                                            AsymmetricPercentileInterval,
                                            ManualInterval, MinMaxInterval)

from astropy.visualization.stretch import (LinearStretch, SqrtStretch,
                                           PowerStretch, LogStretch,
                                           AsinhStretch)

from astropy.visualization.mpl_normalize import ImageNormalize

__all__ = ['simple_norm']


def simple_norm(data, stretch='linear', power=1.0, asinh_a=0.1, log_a=1000,
                min_cut=None, max_cut=None, min_percent=None, max_percent=None,
                percent=None, clip=True):

    if percent is not None:
        interval = PercentileInterval(percent)
    elif min_percent is not None or max_percent is not None:
        interval = AsymmetricPercentileInterval(min_percent or 0.,
                                                max_percent or 100.)
    elif min_cut is not None or max_cut is not None:
        interval = ManualInterval(min_cut, max_cut)
    else:
        interval = MinMaxInterval()

    if stretch == 'linear':
        stretch = LinearStretch()
    elif stretch == 'sqrt':
        stretch = SqrtStretch()
    elif stretch == 'power':
        stretch = PowerStretch(power)
    elif stretch == 'log':
        stretch = LogStretch(log_a)
    elif stretch == 'asinh':
        stretch = AsinhStretch(asinh_a)
    else:
        raise ValueError('Unknown stretch: {0}.'.format(stretch))

    vmin, vmax = interval.get_limits(data)

    return ImageNormalize(vmin=vmin, vmax=vmax, stretch=stretch, clip=clip)
